"""
File Purpose: statistics (e.g. mean, std, min, max)
"""
import warnings

from ..quantity_loader import QuantityLoader
from ...errors import MemorySizeError
from ...tools import (
    simple_property, UNSET,
    xarray_where_finite, xarray_stats, xarray_mean,
    xarray_werrmean, xarray_mean_pm_std,
    xarray_werr2pmstd, xarray_pmstd2werr,
    xarray_werradd, xarray_werrsub, xarray_werrmul, xarray_werrdiv,
    xarray_dims_coords,
)
from ...defaults import DEFAULTS

class StatsLoader(QuantityLoader):
    '''statistics, e.g. mean, std, min, max. Usually applied only along self.stat_dims_for(array).
    By default, stat_dims(array) returns only the self.maindims which appear in the array.
    '''

    # # # "STANDARD" LOGIC / PROPERTIES / HELPER METHODS # # #

    cls_behavior_attrs.register('stat_dims', getdefault=lambda ql: getattr(ql, 'maindims', []))
    @property
    def stat_dims(self):
        '''the dims over which to possibly apply statistics (StatsLoader methods).
        will only apply statics along these dims for an array if they actually appear in the array.
        None --> use self.maindims. (this is the default.)
        See also: self.stat_dims_for(array).
        '''
        if getattr(self, '_stat_dims', None) is None:
            return getattr(self, 'maindims', [])
        else:
            return self._stat_dims
    @stat_dims.setter
    def stat_dims(self, value):
        self._stat_dims = value

    def stat_dims_for(self, array):
        '''return the dims to apply statistics over, for this array.
        Here, returns tuple of d from self.stat_dims if d in array.dims.
        '''
        return tuple(d for d in self.stat_dims if d in array.dims)

    stats_dimpoint_wise = simple_property('_stats_dimpoint_wise', default=None,
        doc='''whether to apply stat calculations at each DimPoint, or after loading the full array.
        [EFF] this setting is just for efficiency; it doesn't affect results (when no MemorySizeError crash).

        True --> apply stat calculations to each array (at each DimPoint) before joining arrays.
                e.g., self('mean_n') gets 'mean_n' at each fluid & snap, then joins results.
        False --> join arrays across all DimPoints before applying stat calculations.
                e.g., self('mean_n') gets 'n' (which varies across fluid & snap), then takes mean.
        None --> if result.size will be > DEFAULTS.STATS_DIMPOINT_WISE_MIN_N / self.get_ncpu(), use True.
                otherwise try using False, then use True if MemorySizeError is raised.
                (True seems to be faster for large arrays but slower for small arrays.
                But also, when ncpu>1, loading across dimpoints is faster due to parallelization.)

        regardless of this setting, stat calculations are applied only along self.stat_dims(array).''')

    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['stats_dimpoint_wise'] + any values from super().
        '''
        # 'stats_dimpoint_wise' doesn't affect numerical results, just efficiency,
        #   which is why it's not in self.behavior_attrs. But, user might still want to set it during __call__.
        return ['stats_dimpoint_wise'] + super()._extra_kw_for_quantity_loader_call

    def _print_or_warn(self, msg, debug_thresh=1):
        '''if DEBUG>=debug_thresh, print msg. Else, warnings.warn(msg).'''
        if DEFAULTS.DEBUG >= debug_thresh:
            print(msg)
        else:
            warnings.warn(msg)

    def _get_stat_var(self, var, statfunc):
        '''returns result of applying statfunc to self(var), possibly one dimpoint at a time then joining results.
        see help(type(self).stats_dimpoint_wise) for details.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if self.stats_dimpoint_wise is None:
            # if result will be a certain size or larger, use True (faster for large arrays).
            ncpu = self.get_ncpu() if hasattr(self, 'get_ncpu') else 1
            min_n = DEFAULTS.STATS_DIMPOINT_WISE_MIN_N
            size_threshold = min_n / ncpu
            if size_threshold is not None:
                try:
                    result_size = self.match_var_result_size(var, maindims=True)
                except Exception:
                    if DEFAULTS.DEBUG >= 1 or getattr(self, 'verbose', False) >= 1:
                        warnmsg = (f'Error when guessing result size for var={var!r}.\n'
                                   f'Could not guess result_dims; maybe self.tree(var={var!r}) crashed.\nUsing default guess: '
                                   f'result_size = self.current_n_dimpoints() * self.maindims_size '
                                   f'== {self.current_n_dimpoints()} * {getattr(self, "maindims_size", 1)}')
                        self._print_or_warn(warnmsg, debug_thresh=1)
                    result_size = self.current_n_dimpoints() * getattr(self, 'maindims_size', 1)
                if result_size > size_threshold:
                    if DEFAULTS.DEBUG >= 1 or getattr(self, 'verbose', False) >= 1:
                        warnmsg = (f'temporarily using stats_dimpoint_wise=True for efficiency; predicted '
                                   f'result_size ({result_size}) > DEFAULTS.STATS_DIMPOINT_WISE_MIN_N (={min_n}) / ncpu (={ncpu}).')
                        self._print_or_warn(warnmsg, debug_thresh=1)
                    return self._get_stat_var_dimpoint_wise_true(var, statfunc)
            # else, try with False first (faster for small arrays). Switch to True if MemorySizeError when False.
            try:
                return self._get_stat_var_dimpoint_wise_false(var, statfunc)
            except MemorySizeError:
                if DEFAULTS.DEBUG >= 1 or getattr(self, 'verbose', False) >= 1:
                    warnmsg = (f'MemorySizeError when avoiding dimpoint_wise stats due to stats_dimpoint_wise=None. '
                               f'Will use stats_dimpoint_wise=True instead. (var={var!r}) '
                               f'(Might want to adjust DEFAULTS.ARRAY_MBYTES_MAX, '
                               f'or explicity set stats_dimpoint_wise=False to see the MemorySizeError crash.)')
                    self._print_or_warn(warnmsg, debug_thresh=1)
                return self._get_stat_var_dimpoint_wise_true(var, statfunc)
        elif self.stats_dimpoint_wise:
            return self._get_stat_var_dimpoint_wise_true(var, statfunc)
        else:
            return self._get_stat_var_dimpoint_wise_false(var, statfunc)

    def _get_stat_var_dimpoint_wise_false(self, var, statfunc):
        '''returns result of applying statfunc to self(var). (Used when self.stats_dimpoint_wise=False.)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        value = self(var)
        return statfunc(value)
    
    def _get_statfunc_of_var(self, var, statfunc):
        '''load this stat, for self(var).'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        value = self(var)
        return statfunc(value)

    def _get_stat_var_dimpoint_wise_true(self, var, statfunc, min_split=UNSET):
        '''returns result of applying statfunc to self(var). (Used when self.stats_dimpoint_wise=True.)
        min_split: UNSET or int
            [EFF] min length of dimension before actually applying dimpointwise stats along it.
            1 --> no minimum. If MemorySizeError when min_split > 1, try again but with min_split=1.
            UNSET --> use DEFAULTS.STATS_DIMPOINT_WISE_MIN_SPLIT.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        def loading_stat():
            '''loads this stat, for self(var), at a single dimpoint.'''
            __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
            var_value = self(var)
            return statfunc(var_value)
        if min_split is UNSET: min_split = DEFAULTS.STATS_DIMPOINT_WISE_MIN_SPLIT
        try:
            value = self.load_across_dims_implied_by(var,
                        self._get_statfunc_of_var, var, statfunc,   # loader & *args_loader
                        _min_split=min_split)
        except MemorySizeError:
            if min_split == 1:
                raise
            # else
            if DEFAULTS.DEBUG >= 5 or getattr(self, 'verbose', False) >= 4:
                warnmsg = (f'MemorySizeError when min_split={min_split}, during stats_dimpoint_wise=True. '
                           f'Trying min_split=1 instead. var={var!r}. (Might want to adjust DEFAULTS.ARRAY_MBYTES_MAX.)')
                if not (DEFAULTS.DEBUG >= 5):
                    warnings.warn(warnmsg)  # only verbose, not DEBUG.
                else:
                    print(warnmsg)  # DEBUG --> print, not warn.
            return self._get_stat_var_dimpoint_wise_true(var, statfunc, min_split=1)
        return value


    # # # "STANDARD" STATS (ACROSS MAINDIMS) # # #

    # 'sum_{var}'
    @known_pattern(r'sum_(.+)', deps=[0])
    def get_sum(self, var, *, _match=None):
        '''sum. sum(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_sum)
    def _stat_dims_sum(self, array):
        return array.sum(dim=self.stat_dims_for(array))

    # 'mean_{var}'
    @known_pattern(r'mean_(.+)', deps=[0])
    def get_mean(self, var, *, _match=None):
        '''mean. mean(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_mean)
    def _stat_dims_mean(self, array):
        return array.mean(dim=self.stat_dims_for(array))

    # 'std_{var}'
    @known_pattern(r'std_(.+)', deps=[0])
    def get_std(self, var, *, _match=None):
        '''standard deviation. std(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_std)
    def _stat_dims_std(self, array):
        return array.std(dim=self.stat_dims_for(array))

    # 'min_{var}'
    @known_pattern(r'min_(.+)', deps=[0])
    def get_min(self, var, *, _match=None):
        '''minimum. min(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_min)
    def _stat_dims_min(self, array):
        return array.min(dim=self.stat_dims_for(array))

    # 'max_{var}'
    @known_pattern(r'max_(.+)', deps=[0])
    def get_max(self, var, *, _match=None):
        '''maximum. max(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_max)
    def _stat_dims_max(self, array):
        return array.max(dim=self.stat_dims_for(array))

    # 'median_{var}'
    @known_pattern(r'median_(.+)', deps=[0])
    def get_median(self, var, *, _match=None):
        '''median. median(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_median)
    def _stat_dims_median(self, array):
        return array.median(dim=self.stat_dims_for(array))

    # 'rms_{var}'
    @known_pattern(r'rms_(.+)', deps=[0])
    def get_rms(self, var, *, _match=None):
        '''root mean square. sqrt(mean(var**2)). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_rms)
    def _stat_dims_rms(self, array):
        return (array**2).mean(dim=self.stat_dims_for(array))**0.5

    # 'delta_{var}'
    @known_pattern(r'delta_(.+)', deps=[0])
    def get_delta(self, var, *, _match=None):
        '''perturbation. var - mean(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_delta)
    def _stat_dims_delta(self, array):
        return array - array.mean(dim=self.stat_dims_for(array))

    # 'deltafrac_{var}'
    @known_pattern(r'deltafrac_(.+)', deps=[0])
    def get_deltafrac(self, var, *, _match=None):
        '''perturbation. (var - mean(var)) / mean(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_deltafrac)
    def _stat_dims_deltafrac(self, array):
        return (array / array.mean(dim=self.stat_dims_for(array))) - 1  # == (array - mean) / mean

    # 'meannormed_{var}'
    @known_pattern(r'meannormed_(.+)', deps=[0])
    def get_meannormed(self, var, *, _match=None):
        '''normalized by mean. var / mean(var). Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_meannormed)
    def _stat_dims_meannormed(self, array):
        return array / array.mean(dim=self.stat_dims_for(array))

    # 'weighted_{weights}_mean_{var}'
    @known_pattern(r'weighted_(.+)_mean_(.+)', deps=[0, 1])
    def get_weighted_mean(self, weights_var, *, _match=None):
        '''mean, weighted by weights. mean(weights*var)/mean(weights).
        E.g. 'weighted_n_mean_T' --> mean(n * T) / mean(n).
            (see also: 'nmean_[var]', which is a shorthand for 'weighted_n_mean_[var]')
        Applied only along any self.stat_dims in array.
        '''
        weights, var = _match.groups()
        return self(f'mean_({var}*meannormed_{weights})')  # mean(T*n/mean(n)) equivalent to mean(n*T)/mean(n).

    # 'weighted_{weights}_std_{var}'
    @known_pattern(r'weighted_(.+)_std_(.+)', deps=[0, 1])
    def get_weighted_std(self, weights_var, *, _match=None):
        '''std, weighted by weights. std(weights*var)/mean(weights).
        E.g. 'weighted_n_std_mod_u' --> std(n * mod_u) / mean(n).
            (see also: 'nstd_[var]', which is a shorthand for 'weighted_n_std_[var]')
        Applied only along any self.stat_dims in array.

        Currently, equivalent to self('(std_({weights}*{var}))/mean(weights)')
        [TODO][EFF] internally, don't compute weights twice...
        '''
        weights, var = _match.groups()
        return self(f'(std_({weights}*{var}))/mean_{weights}')

    # 'stats_{var}'
    @known_pattern(r'stats_(.+)', deps=[0])
    def get_stats(self, var, *, _match=None):
        '''return dataset of stats for var
        stats include: mean, std, min, max, median, rms.
        Applied only along any self.stat_dims in array.

        Incompatible with Dataset vars.
        Consider also: self('astats_var'), self('var').pc.stats()
        '''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_stats)
    def _stat_dims_stats(self, array):
        return xarray_stats(array, dim=self.stat_dims_for(array))

    @known_pattern(r'astats_(.+)', deps=[0])
    def get_astats(self, var, *, _match=None):
        '''return dataarray of stats for var, reporting stats along new dim: 'stat'.
        stats include: mean, std, min, max, median, rms.
        Applied only along any self.stat_dims in array.

        Compatible with Dataset vars (without existing 'stat' coord or dim).
        The result excludes whichever dims the stats are being taken across,
            and adds the new dim 'stat' with the stats.

        Consider also: self('stats_var'), self('var').pc.stats(to_da='stat')
        '''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_astats)
    def _stat_dims_astats(self, array):
        return xarray_stats(array, dim=self.stat_dims_for(array), to_da='stat')

    # # # FINITE VALUES ONLY - e.g. "nanmean" # # #

    @known_pattern(r'finite_(.+)', deps=[0])    # 'finite_{var}'
    def get_finite(self, var, *, _match=None):
        '''var, masked with NaN wherever values are not finite.'''
        here, = _match.groups()
        value = self(here)
        return xarray_where_finite(value)

    @known_pattern(r'nanmean_(.+)', deps=[0])    # 'nanmean_{var}'
    def get_nanmean(self, var, *, _match=None):
        '''mean. mean(var), ignoring NaNs AND infs. Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self(f'mean_finite_{here}')

    @known_pattern(r'nanstd_(.+)', deps=[0])    # 'nanstd_{var}'
    def get_nanstd(self, var, *, _match=None):
        '''standard deviation. std(var), ignoring NaNs AND infs. Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self(f'std_finite_{here}')

    @known_pattern(r'nanmin_(.+)', deps=[0])    # 'nanmin_{var}'
    def get_nanmin(self, var, *, _match=None):
        '''minimum. min(var), ignoring NaNs AND infs. Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self(f'min_finite_{here}')

    @known_pattern(r'nanmax_(.+)', deps=[0])    # 'nanmax_{var}'
    def get_nanmax(self, var, *, _match=None):
        '''maximum. max(var), ignoring NaNs AND infs. Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self(f'max_finite_{here}')

    @known_pattern(r'nanmedian_(.+)', deps=[0])    # 'nanmedian_{var}'
    def get_nanmedian(self, var, *, _match=None):
        '''median. median(var), ignoring NaNs AND infs. Applied only along any self.stat_dims in array.'''
        here, = _match.groups()
        return self(f'median_finite_{here}')

    @known_pattern(r'nanrms_(.+)', deps=[0])    # 'nanrms_{var}'
    def get_nanrms(self, var, *, _match=None):
        '''root mean square. sqrt(mean(var**2)), ignoring NaNs AND infs.
        Applied only along any self.stat_dims in array.
        '''
        here, = _match.groups()
        return self(f'rms_finite_{here}')

    @known_pattern(r'nandelta_(.+)', deps=[0])    # 'nandelta_{var}'
    def get_nandelta(self, var, *, _match=None):
        '''perturbation. var - mean(var), ignoring NaNs AND infs.
        Applied only along any self.stat_dims in array.
        '''
        here, = _match.groups()
        return self(f'delta_finite_{here}')

    @known_pattern(r'nandeltafrac_(.+)', deps=[0])    # 'nandeltafrac_{var}'
    def get_nandeltafrac(self, var, *, _match=None):
        '''perturbation. (var - mean(var)) / mean(var), ignoring NaNs AND infs.
        Applied only along any self.stat_dims in array.
        '''
        here, = _match.groups()
        return self(f'deltafrac_finite_{here}')


    # # # ERROR PROPAGATION / MATH # # #

    cls_behavior_attrs.register('werrmath_require_std', default=False)
    werrmath_require_std = simple_property('_werrmath_require_std', default=False,
        doc='''whether to require 'std' data var appear in at least 1 input for all werrmath operations.
        E.g. if True and doing A_werradd_B but neither A nor B has 'std', crash with InputError.''')

    @known_pattern(r'werrmean_(.+)', deps=[0])   # 'werrmean_{var}'
    def get_werrmean(self, var, *, _match=None):
        '''dataset of 'mean' and 'std' of var. Computed along self.stat_dims in array.'''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_werrmean)
    def _stat_dims_werrmean(self, array):
        return xarray_werrmean(array, dim=self.stat_dims_for(array))

    @known_pattern(r'mean_pm_std_(.+)', deps=[0])   # 'mean_pm_std_{var}'
    def get_mean_pm_std(self, var, *, _match=None):
        '''return dataset of 'mean+std', 'mean', 'mean-std' for var.
        Computed along any self.stat_dims in array.

        Equivalent: werr2pmstd_werrmean_var
        '''
        here, = _match.groups()
        return self._get_stat_var(here, self._stat_dims_mean_pm_std)
    def _stat_dims_mean_pm_std(self, array):
        return xarray_mean_pm_std(array, dim=self.stat_dims_for(array))

    @known_pattern(r'werr2pmstd_(.+)', deps=[0])   # 'werr2pmstd_{var}'
    def get_werr2pmstd(self, var, *, _match=None):
        '''convert dataset with 'mean' and 'std' into dataset with 'mean+std', 'mean', 'mean-std'.
        werr2pmstd_var will crash if self(var) doesn't have 'mean' and 'std' data vars.
        '''
        here, = _match.groups()
        werr_val = self(here)
        return xarray_werr2pmstd(werr_val)

    @known_pattern(r'pmstd2werr_(.+)', deps=[0])   # 'pmstd2werr_{var}'
    def get_pmstd2werr(self, var, *, _match=None):
        '''convert dataset with 'mean+std' and 'mean-std' into dataset with 'mean' and 'std'.
        pmstd2werr_var will crash if self(var) doesn't have 'mean+std' and 'mean-std' data vars.
        '''
        here, = _match.groups()
        pmstd_val = self(here)
        return xarray_pmstd2werr(pmstd_val)

    @known_pattern(r'(.+)_werradd_(.+)', deps=[0, 1])   # '{var}_werradd_{var2}'
    def get_werradd(self, var, *, _match=None):
        '''A_werradd_B = A + B, but result is a dataset with 'mean' and 'std'.
        Does not take any means or stds here, but if A or B has 'std' already,
        assumes independent errors and applies "error propagation" formula:
            mean(A + B) = mean(A) + mean(B)
            std(A + B) = sqrt(std(A)**2 + std(B)**2)
        (if A or B is DataArray, treat as 'mean'. if missing 'std', assume 0.)
        '''
        A, B = _match.groups()
        Aval = self(A)
        Bval = self(B)
        return xarray_werradd(Aval, Bval, require_std=self.werrmath_require_std)

    @known_pattern(r'(.+)_werrsub_(.+)', deps=[0, 1])   # '{var}_werrsub_{var2}'
    def get_werrsub(self, var, *, _match=None):
        '''A_werrsub_B = A - B, but result is a dataset with 'mean' and 'std'.
        Does not take any means or stds here, but if A or B has 'std' already,
        assumes independent errors and applies "error propagation" formula:
            mean(A - B) = mean(A) - mean(B)
            std(A - B) = sqrt(std(A)**2 + std(B)**2)
        (if A or B is DataArray, treat as 'mean'. if missing 'std', assume 0.)
        '''
        A, B = _match.groups()
        Aval = self(A)
        Bval = self(B)
        return xarray_werrsub(Aval, Bval, require_std=self.werrmath_require_std)

    @known_pattern(r'(.+)_werrmul_(.+)', deps=[0, 1])   # '{var}_werrmul_{var2}'
    def get_werrmul(self, var, *, _match=None):
        '''A_werrmul_B = A * B, but result is a dataset with 'mean' and 'std'.
        Does not take any means or stds here, but if A or B has 'std' already,
        assumes independent errors and applies "error propagation" formula:
            z = A * B
            mean(z) = mean(A) * mean(B)
            std(z) = abs(mean(z)) * sqrt((std(A)/mean(A))**2 + (std(B)/mean(B))**2)
        (if A or B is DataArray, treat as 'mean'. if missing 'std', assume 0.)
        '''
        A, B = _match.groups()
        Aval = self(A)
        Bval = self(B)
        return xarray_werrmul(Aval, Bval)

    @known_pattern(r'(.+)_werrdiv_(.+)', deps=[0, 1])   # '{var}_werrdiv_{var2}'
    def get_werrdiv(self, var, *, _match=None):
        '''A_werrdiv_B = A / B, but result is a dataset with 'mean' and 'std'.
        Does not take any means or stds here, but if A or B has 'std' already,
        assumes independent errors and applies "error propagation" formula:
            z = A / B
            mean(z) = mean(A) / mean(B)
            std(z) = abs(mean(z)) * sqrt((std(A)/mean(A))**2 + (std(B)/mean(B))**2)
        (if A or B is DataArray, treat as 'mean'. if missing 'std', assume 0.)
        '''
        A, B = _match.groups()
        Aval = self(A)
        Bval = self(B)
        return xarray_werrdiv(Aval, Bval)


    # # # MISC # # #
    # stats but not applied across maindims.

    @known_pattern(r'meant_(.+)', deps=[0])
    def get_meant(self, var, *, _match=None):
        '''mean-across-time of var.
        self(meant_var) --> self(var).mean(tdim), where tdim is the dim associated with time,
                (tdim default 'snap', but if 't' coord associated with a dim, use that dim.)
        '''
        here, = _match.groups()
        val = self(here)
        tdim = 'snap'  # default tdim
        dims_coords = xarray_dims_coords(val)
        for k, coords in dims_coords.items():
            if 't' in coords:
                tdim = k
        return xarray_mean(val, tdim, missing_dims='ignore')

    @known_pattern(r'werrmeant_(.+)', deps=[0])
    def get_werrmeant(self, var, *, _match=None):
        '''dataset of 'mean' and 'std' of var, taking stats across time dimension.
        self(werrmeant_var) --> self(var).werrmean(tdim), where tdim is the dim associated with time,
                (tdim default 'snap', but if 't' coord associated with a dim, use that dim.)
        '''
        here, = _match.groups()
        val = self(here)
        tdim = 'snap'
        dims_coords = xarray_dims_coords(val)
        for k, coords in dims_coords.items():
            if 't' in coords:
                tdim = k
        return xarray_werrmean(val, tdim, missing_dims='ignore')



"""
File Purpose: binding functions to already-existing classes

Motivation: it is convenient to be able to add many functions to classes after their creation.
This facilitates a useful design principle:
    - when creating a class, only define a minimal set of behaviors.
    - when adding new behaviors to an existing class, bind new attributes/methods to the class,
        OUTSIDE of the class definition, rather than editing the class defintion directly.


Usually it is sufficient to use this module in the following way:
    from binding_module import Binding
    binding = Binding(locals())

    class MyClass(...):
        # this class could be defined here, or literally anywhere else.
        # e.g. "from module_with_class import MyClass" is fine too.

    with binding.to(MyClass):
        @binding
        def foo1(self, *args, **kw):    # any signature is fine, e.g. foo1(self, x) is fine too.
            # < code for foo1 goes here

        @binding
        def foo2(self, *args, **kw):
            # < code for foo2 goes here

        # ... can define any number of functions using @binding

    print(MyClass.foo1)
    --> (info about the bound method foo1 of MyClass, defined above)
    print(foo1)
    --> (NameError; foo1 is undefined in this namespace after exiting the 'with' block.)

It is also possible to keep a local copy(s) to the defined function(s),
or to bind them as staticmethod or classmethod. See help(Binding) for details.
"""

from ..docs_tools import format_docstring
from ..properties import (
    alias, alias_to_result_of,
)
from ..sentinels import NO_VALUE
from ...errors import BindingError


''' --------------------- Lightweight single-use decorator --------------------- '''

def bind_to(*targets):
    '''returns a function decorator which binds f to target.(f.__name__) for target in targets.
    This decorator returns the unbound version of f.
    Example, equivalent to defining foo inside MyClass, and also in the local namespace:
        @bind_to(MyClass)
        def foo(self, args):
            # <code for foo>

    If you want foo to no longer be defined in the local namespace, you can delete it afterwards:
        del foo

    If you want to do this for many functions but don't want to delete each one manually,
    consider using Binding() instead.
    '''
    def bind_then_return_f(f):
        '''binds f to targets, then returns f.'''
        for target in targets:
            setattr(target, f.__name__, f)
        return f
    return bind_then_return_f


''' --------------------- the Binding class --------------------- '''

class Binding():
    '''context manager for binding environment.
    provides these function decorators:
        @self (or @self.bind()) --> bind function or class to each target from self.default_targets.
        @self.binder(*targets) --> bind function or class to each target from targets.
    bindings at the corresponding attribute name e.g. target.(func.__name__) = func.
    Upon exiting, restore original state for names in self.namespace of any objects
        decorated via @self, @self.bind(), or @self.bind_to().

    To bind, but ALSO keep the function in the local namespace (i.e. don't remove it upon exiting),
        use keep_local=True. Either in 'self.bind(keep_local=True)'
        or 'with binding.to(*targets, keep_local=True)'.
    To bind a special method type (e.g., staticmethod, classmethod), use the methodtype parameter,
        e.g. methodtype=staticmethod. Either in 'self.bind(methodtype=staticmethod)'
        or 'with binding.to(..., methodtype=staticmethod)'
    Also, upon exiting context, (i.e., after 'with' block),
        restore self.keep_local and self.methodtype to their original values
        (or their values just before the most recent call to self.to(...),
        if context hasn't been exited since last call to self.to.)

    EXAMPLE:
    # originally, foo1, Foo3, foo4, foo5 undefined, but foo2 defined as foo2_original.
    binding = Binding(locals())
    with binding.to(MyClass):
        @binding
        def foo1(self, args):          # binds MyClass.foo1 = foo1
            # <code for foo1>

        @binding.binder(MySecondClass)
        def foo2(self, args):          # binds MySecondClass.foo2 = foo2
            # <code for foo2>

        @binding.bind()  # equivalent to @binding
        class Foo3(self, args):        # binds MyClass.Foo3 = Foo3
            # <code for Foo3>

        @binding.binder(MySecondClass, MyThirdClass)
        def foo4(self, args):          # binds MySecondClass.foo4 = foo4; MyThirdClass.foo4 = foo4
            # <code for foo4>

        @binding.bind(keep_local=True)
        def foo5(self, args):          # MyClass.foo5 = foo5
            # <code for foo5>

        @binding.bind(methodtype=staticmethod)
        def foo6(args):                # MyClass.foo6 = staticmethod(foo6)
            # <code for foo6>

    # upon exiting, alter the local namespace: delete foo1, Foo3, foo4; set foo2 = foo2_original.
    # (do not delete foo5 from the namespace since keep_local=True was used for foo5.)

    ARGS:
        namespace: dict
            the namespace to clean up upon exiting the context.
            This class was designed with the intention that:
                namespace=locals().
        *default_targets: classes
            the targets to which objects decorated with @self.bind() will be bound
            If this list is empty, trying to use self.bind() will raise a ValueError.
            To enter custom targets, use self.bind_to instead.
        keep_local: bool, default False
            default value for whether to keep local copies of the bound functions.
            E.g. if self.keep_local=True, then by default we will keep local copies.
    '''
    # # # CREATION # # #
    def __init__(self, namespace, *default_targets, keep_local=False, methodtype=None):
        self.namespace = namespace
        self.default_targets = default_targets
        self.keep_local = keep_local
        self.methodtype = methodtype
        self._remember_settings = dict()  # << restore values then clear upon exiting 'with' block.
        self._reset_flagged()

    # # # CONTEXT MANAGER # # #
    def _reset_flagged(self):
        '''resets/creates flagged trackers in self.'''
        self.flagged_for_restoration = dict()
        self.flagged_for_deletion = set()
        
    def __enter__(self):
        memory = self._remember_settings
        if 'keep_local' not in memory:
            memory['keep_local'] = self.keep_local
        if 'methodtype' not in memory:
            memory['methodtype'] = self.methodtype

    def __exit__(self, *_unused_args):
        self.cleanup()
        for attr, value in self._remember_settings.items():
            setattr(self, attr, value)
        self._remember_settings.clear()

    def cleanup(self):
        '''cleans up self.namespace appropriately, following these rules:
        1) previously-defined names in self.namespace (defined first without using self)
            which were overwritten to point at an object wrapped by self
            will be restored to their original values.
        2) previously undefined objects in self.namespace
            which were written to point at an object wrapped by self
            will be removed from self.namespace.
        '''
        for name in self.flagged_for_deletion:
            del self.namespace[name]
        for fname, forig in self.flagged_for_restoration.items():
            self.namespace[fname] = forig
        self._reset_flagged()

    # # # BINDING # # #
    _bind_kwarg_docs = \
        '''keep_local: bool or None, default None
            if None, use self.keep_local instead. (self.keep_local=False by default)
            if True, instead don't flag for cleanup, and return the original function.
            else (default), flag for cleanup, and return a BindingError object
                    (which is only relevant if attempting to access the function before cleanup).
        methodtype: None, staticmethod, or classmethod
            if None, use self.methodtype instead. (self.methodtype=None by default)
            if still not None, bind methodtype(f) instead of f.'''

    @format_docstring(_bind_kwarg_docs=_bind_kwarg_docs)
    def bind(self, *, keep_local=None, methodtype=None):
        '''returns decorator; decorator(f) binds f to target.(f.__name__) for target in self.default_targets,
        and also flags f appropriately for cleanup later when exiting the Binding context.

        kwargs determine the behavior for the resulting decorator:
        {_bind_kwarg_docs}
        '''
        if keep_local is None: keep_local = self.keep_local
        if len(self.default_targets) == 0:
            raise ValueError('self.bind() uses self.default_targets, but self.default_targets is empty!')
        return self.binder(*self.default_targets, keep_local=keep_local, methodtype=methodtype)

    @format_docstring(_bind_kwarg_docs=_bind_kwarg_docs)
    def binder(self, *targets, keep_local=None, methodtype=None):
        '''returns decorator(f); decorator(f) binds f to target.(f.__name__) for target in targets,
        and also flags f appropriately for cleanup later when exiting the Binding context.

        if keep_local and methodtype determine behavior for the resulting decorator_maker:
        {_bind_kwarg_docs}
        '''
        if keep_local is None:
            keep_local = self.keep_local
        if methodtype is None:
            methodtype = self.methodtype
        kw = dict(keep_local=keep_local, methodtype=methodtype)
        # Could define the result with 'if' statements inside it,
        #   but it's easier to debug anything going wrong if there are
        #   different function names, based on keep_local and methodtype.
        # use_result is shorthand so we don't need to write all the function names twice.
        result = dict()   # easier for me to understand than 'nonlocal' or 'global'
        def use_result(decorator):
            result['val'] = decorator
            return decorator
        # loop through all combinations of keep_local and methodtype.
        if methodtype is None:
            if keep_local:
                @use_result
                def bind_and_keep_f(f): return self.direct_bind(f, *targets, **kw)
            else:
                @use_result
                def bind_f(f): return self.direct_bind(f, *targets, **kw)
        elif methodtype is staticmethod:
            if keep_local:
                @use_result
                def bind_staticmethod_and_keep_f(f): return self.direct_bind(f, *targets, **kw)
            else:
                @use_result
                def bind_staticmethod_f(f): return self.direct_bind(f, *targets, **kw)
        elif methodtype is classmethod:
            if keep_local:
                @use_result
                def bind_classmethod_and_keep_f(f): return self.direct_bind(f, *targets, **kw)
            else:
                @use_result
                def bind_classmethod_f(f): return self.direct_bind(f, *targets, **kw)
        else:
            if keep_local:
                @use_result
                def bind_specialmethod_and_keep_f(f): return self.direct_bind(f, *targets, **kw)
            else:
                @use_result
                def bind_specialmethod_f(f): return self.direct_bind(f, *targets, **kw)
        return result['val']

    @format_docstring(_bind_kwarg_docs=_bind_kwarg_docs)
    def direct_bind(self, f, *targets, keep_local=None, methodtype=None):
        '''binds f to target.(f.__name__) for target in targets, then flags f appropriately for cleanup.
        named "direct" bind because this is not a function decorator.

        {_bind_kwarg_docs}
        '''
        if keep_local is None:
            keep_local = self.keep_local
        if methodtype is None:
            methodtype = self.methodtype
        fname = f.__name__
        # bind to targets
        fbind = f if (methodtype is None) else methodtype(f)
        for target in targets:
            setattr(target, fname, fbind)
        # handle flagging appropriately for cleanup later.
        if not keep_local:
            restore = self.flagged_for_restoration
            delete = self.flagged_for_deletion
            if (fname in restore) or (fname in delete):
                return   # already flagged; don't flag again.
            if fname in self.namespace:
                restore[fname] = self.namespace[fname]
            else:
                self.flagged_for_deletion = delete.union((fname,))
        # return result
        return f if keep_local else BindingError(f)

    @format_docstring(_bind_kwarg_docs=_bind_kwarg_docs)
    def bind_to(self, cls, f, *, keep_local=None, methodtype=None):
        '''bind f to cls. Equivalent to self.direct_bind(self, f, cls, **kw) with same kw as provided here.

        {_bind_kwarg_docs}
        '''
        return self.direct_bind(f, cls, keep_local=keep_local, methodtype=methodtype)

    # # # FREQUENTLY USED CONVENIENCE # # #
    def to(self, *targets, keep_local=NO_VALUE, methodtype=NO_VALUE):
        '''sets targets in self then returns self.
        convenient for writing context entry like this:
        binding = Binding(locals())
        with binding.to(MyClass1):
            @binding
            def foo1(args):
                #...

        if keep_local and/or methodtype are provided,
            set those attributes in self, but ALSO provide the instruction to self
            to return those attributes to their previous state upon next exiting context.
        '''
        self.default_targets = targets
        memory = self._remember_settings
        if keep_local is not NO_VALUE:
            memory['keep_local'] = self.keep_local
            self.keep_local = keep_local
        if methodtype is not NO_VALUE:
            memory['methodtype'] = self.methodtype
            self.methodtype = methodtype
        return self

    targetting = alias('to')

    __call__ = alias_to_result_of('bind',
        doc='''alias to result of self.bind().
        this means you can write @self instead of @self.bind(). E.g:
        binding = Binding(locals())
        with binding.to(MyClass1):
            @binding
            def foo1(args):
                #...
        ''')

    # # # LESS-OFTEN USED CONVENIENCE # # #
    def with_targets(self, *targets):
        '''return Binding object with same namespace as self, but using the provided targets instead.'''
        return Binding(self.namespace, *targets)

    with_target = alias('with_targets')

    target = property(lambda self: self.get_target(),
            lambda self, value: self.set_target(value),
            doc='''@self.bind() binds objects to target.(object.__name__).
            When len(self.default_targets)==0, roughly an alias to self.default_targets.
            Otherwise irrelevant; see self.default_targets instead.''')

    def get_target(self):
        '''gets binding target of self. makes ValueError if self has not precisely 1 target.'''
        if len(self.default_targets) == 1:
            return self.default_targets[0]
        else:
            raise ValueError(f'zero or multiple targets detected: {self.default_targets}')

    def set_target(self, target):
        '''sets binding target of self to target. (used in self.bind)'''
        self.default_targets = [target]



"""
File Purpose: tools related to cross sections for collisions
"""

import ast
import os

import numpy as np
import xarray as xr

from ...errors import FileContentsError, InputError
from ...units import UnitsManager
from ...tools import (
    simple_property, alias,
    SymmetricPairMapping,
)

from ...defaults import DEFAULTS

# location of the cross_section_tables folder, on this machine.
# this is used when trying to get one of the default cross section tables.
# __file__ means the file where this line of code is located.
DEFAULTS.CROSS_SECTIONS_DIR = os.path.join(os.path.dirname(__file__), 'cross_section_tables')


class UnitsManagerCrossUnits(UnitsManager):
    '''UnitsManager for cross sections, since crossunits are in [cm^-2].'''
    @classmethod
    def from_crossunits(cls, crossunits, **kw):
        '''return UnitsManagerCrossUnits with length units conversions based on crossunits.
            cross_section [si] = result('area', 'si') * cross_section [from file]
        equivalent to cls(l=np.sqrt(crossunits) * 1e-2, **kw)
        '''
        l = np.sqrt(crossunits) * 1e-2   # the 1e-2 converts cm to m
        return cls(l=l, **kw)


''' --------------------- CrossTable, and getting from file --------------------- '''

class CrossTable():
    '''stores data about a cross sections table.

    kT: temperatures [eV]. array of values
    cross: cross sections [crossunits]. array of values
    crossunits: conversion factor; cross * crossunits = cross section [cm^-2]. single value.
    fc  numerical factor, based on convention of the cross table.
        when getting the collision frequency, the formula needs to use fc * cross.
        PlasmaCalcs 'collisions_cross_section' ('collcross') already includes the fc factor.
        The value of fc should probably be set as follows:
            tables utilizing Bruno+2010 should have fc=4/3.
            tables utilizing Vranjes+2013 should have fc=1.
        By default, should prefer to use Bruno+2010 (hence the default is 4/3),
            as explained in Wargnier+2022.

    note: some cross tables are available by default;
        these can be loaded with cls.default(name), where name is one of cls.DEFAULTS.keys().
    '''
    # # # INITIALIZATION / CREATION # # #
    def __init__(self, kT, cross, crossunits, fc=4/3):
        self.kT = kT
        self.cross = cross
        self.crossunits = crossunits
        self.fc = fc
        self.u = UnitsManagerCrossUnits.from_crossunits(crossunits)

    @classmethod
    def from_file(cls, filename, fc=4/3):
        '''return CrossTable from file. Be sure to put the appropriate value of fc.
        Bruno+2010 should have fc=4/3. Vranjes+2013 should have fc=1.
        see help(CrossTable) for more fc info. 
        '''
        cross_data = read_cross_text(filename)
        result = cls(**cross_data, fc=fc)
        result.filename = os.path.abspath(filename)
        result.dirname = os.path.dirname(result.filename)
        result.filebase = os.path.basename(result.filename)
        return result

    @classmethod
    def from_defaults(cls, name):
        '''return CrossTable from one of the default cross table files.
        name = one of cls.DEFAULTS.keys().
        '''
        filename, fc = cls._default_file(name)
        return cls.from_file(filename, fc=fc)

    DEFAULTS = {   # dict of {shorthand: (filename, fc)}
        # shorthand: e=electrons; p=protons; h=H(neutral); he=He(neutral); hep=He+; hepp=He++
        # order doesn't really matter, but grouping files here for better readability...
        # -- bruno -- charged-neutral collisions --
        'e-h'    : ('e-h-bruno-fits.txt',     4/3),
        'e-he'   : ('e-he-bruno-fits.txt',    4/3),
        'h-p'    : ('h-p-bruno-fits.txt',     4/3),
        'h-hep'  : ('h-hep-bruno-fits.txt',   4/3),
        'h-hepp' : ('h-hepp-bruno-fits.txt',  4/3),
        'he-p'   : ('he-p-bruno-fits.txt',    4/3),
        'he-hep' : ('he-hep-bruno-fits.txt',  4/3),
        'he-hepp': ('he-hepp-bruno-fits.txt', 4/3),
        # -- bruno -- neutral-neutral collisions --
        'h-h'    : ('h-h-bruno-fits.txt',     4/3),
        'he-h'   : ('he-h-bruno-fits.txt',    4/3),
        'he-he'  : ('he-he-bruno-fits.txt',   4/3),
        # -- vranjes -- charged-neutral collisions --
        'e-h_vranjes'  : (os.path.join('vranjes', 'e-h-cross.txt'), 1),
        'e-he_vranjes' : (os.path.join('vranjes', 'e-he-cross.txt'), 1),
        'h-p_vranjes'  : (os.path.join('vranjes', 'p-h-cross.txt'), 1),
        'he-p_vranjes' : (os.path.join('vranjes', 'p-he-cross.txt'), 1),
        # -- vranjes -- neutral-neutral collisions --
        'h-h_vranjes'  : (os.path.join('vranjes', 'h-h-cross.txt'), 1),
        'he-he_vranjes': (os.path.join('vranjes', 'he-he-cross.txt'), 1),
    }

    @classmethod
    def _default_file(cls, name):
        '''return (abspath to one of the default cross table files, fc for that file)
        name = one of cls.DEFAULTS.keys().
        '''
        try:
            filename, fc = cls.DEFAULTS[name]
        except KeyError:  # make a nicer error message in this case.
            raise InputError(f'name={name!r}; expected one of {tuple(cls.DEFAULTS.keys())}.') from None
        return (os.path.join(DEFAULTS.CROSS_SECTIONS_DIR, filename), fc)

    # # # PROPERTIES / DATA # # #
    T = simple_property('_T', setdefaultvia='_get_T', doc='''temperatures [K], associated with kT.''')

    def _get_T(self):
        '''return T based on self.kT'''
        return self.kT * self.u('eV kB-1', 'si')

    @property
    def cross_si(self):
        '''cross sections, in [m^-2].'''  # [TODO][EFF] caching?
        return self.cross * self.u('area', 'si')

    @property
    def collcross_si(self):
        '''cross sections, in [m^-2], multiplied by fc.'''
        return self.cross_si * self.fc

    def lims(self, attr):
        '''return self.attr.min(), self.attr.max()'''
        val = getattr(self, attr)
        return val.min(), val.max()

    def keys(self):
        '''returns keys of self, as a list.
        These are: ['kT', 'cross', 'cross_si', 'crossunits', 'T']
        '''
        return ['kT', 'cross', 'cross_si', 'crossunits', 'T']

    def __getitem__(self, key):
        '''returns self.key, if key in self.keys().'''
        keys = self.keys()
        if key in keys:
            return getattr(self, key)
        else:
            raise KeyError(f'key={key!r} not recognized; expected one of {keys}')

    # # # INTERPOLATING # # #
    def interp(self, value, input='T', output='collcross_si', *, log=True, **kw_interp):
        '''interpolate value from input (default: 'T') to output (default: 'collcross_si')
        returns result like value (e.g., number if number; array if array),
            with numerical values corresponding to output variable instead of input variable.

        value: number or array-like
            values to interpolate
            note: if value is an xarray, return an xarray with same dims, attrs, and coords as value.
        input: str
            input variable; should be an attribute of self.
            value is a value of this variable.
        output: str
            output variable; should be an attribute of self.
            result is a value of this variable.
        log: bool, default True
            whether to interpolate on log scale.
            True --> interpolate log10(value) against log10(input).
        additional kwargs go to np.interp.

        example: interp(2000, 'T', 'cross_si') --> single value, with cross section at T=2000 K.
        '''
        # get input values
        input_vals = getattr(self, input)
        output_vals = getattr(self, output)
        # interpolate
        if log:
            input_vals = np.log10(input_vals)
            value = np.log10(value)
        result = np.interp(value, input_vals, output_vals, **kw_interp)
        if isinstance(value, xr.DataArray):  # convert back to xarray if needed!
            result = xr.DataArray(result, dims=value.dims, attrs=value.attrs, coords=value.coords)
        return result

    __call__ = alias('interp')

    # # # CONVENIENT # # #
    def __len__(self):
        return len(self.kT)

    # # # DISPLAY # # #
    def __repr__(self):
        Tmin, Tmax = self.lims('T')
        cmin, cmax = self.lims('cross_si')
        contents = [f'len={len(self)}',
                    f'T min={Tmin:.2e}, max={Tmax:.2e} [K]',
                    f'cross min={cmin:.2e}, max={cmax:.2e} [m^-2]',
                    f'crossunits={self.crossunits:.2e} cm^-2',
                    f'fc={self.fc:.2f}']
        cstr = ' | '.join(contents)
        return f'{type(self).__name__}({cstr})'

    def __str__(self):
        filebase = getattr(self, 'filebase', None)
        if filebase is not None:
            return f'{type(self).__name__}({filebase!r})'
        else:
            Tmin, Tmax = self.lims('T')
            cmin, cmax = self.lims('cross_si')
            contents = [f'{Tmin:.1e} < T < {Tmax:.1e} K', f'{cmin:.1e} < cross < {cmax:.1e} m^-2']
            cstr = ', '.join(contents)
            return f'{type(self).__name__}({cstr})'


def read_cross_text(filename):
    '''reads cross sections from text file.
    returns a dictionary with keys 'kT', 'cross'. 
        kT = Temperature [eV]. array of values
        cross = cross section [crossunits]. array of values
        crossunits = conversion factor; cross * crossunits = cross section [cm^-2]. single value.

    raises FileContentsError if file is invalid.

    Text file should look like:
        ; any number of blank lines or lines with comments (starting with ';'),
        ;    at any point in the file, are permitted.
        ; ALWAYS ignores any portion of a line after a ';'

        1.0e-16   ; a single number. this is the "crossunits".
        ; Cross sections in this file represent value [cm^-2] = (value in file) * crossunits

        ; E (eV)      (ignored values)    cross section [crossunits]
           0.100        -1.0000           33.873
           0.297        -1.0000           33.350
           0.801        -1.0000           26.996
           1.306        -1.0000           22.475
           1.647        -1.0000           20.092
           1.988        -1.0000           17.587

        ; ... etc. Values above are just an example
        ; any nonzero amount of whitespace between values is allowed.

        ; Extra values (more than 3 values per line) are allowed, but ignored if provided.
    '''
    result = dict(kT=[], cross=[])
    with open(filename, 'r') as f:
        lines = f.readlines()
    _absfile = os.path.abspath(filename)  # useful for error messages / debugging
    # sanitize lines
    lines = [line.split(';', maxsplit=1)[0] for line in lines]  # remove comments
    lines = [line.strip() for line in lines]  # remove whitespace
    lines = [line.split() for line in lines]  # split into words
    # loop through lines, take relevant lines.
    for _i, line in enumerate(lines):   # track i_ to make debugging easier, if there's a crash.
        if len(line) == 0:
            continue  # ignore empty lines
        elif len(line) == 1:
            if 'crossunits' in result:
                errmsg = f'Line {_i} has exactly 1 value in it, but crossunits was already found! file={_absfile!r}'
                raise FileContentsError(errmsg)
            else:
                result['crossunits'] = ast.literal_eval(line[0])
        elif len(line) == 2:
            errmsg = f'Line {_i} has exactly 2 values in it; expected 0, 1, or >2 values. file={_absfile!r}'
            raise FileContentsError(errmsg)
        else:   # len(line) >= 3:
            kT = ast.literal_eval(line[0])
            # ignore line[1]
            cross = ast.literal_eval(line[2])
            # ignore line[3 or more]
            result['kT'].append(kT)
            result['cross'].append(cross)
    # convert to numpy arrays
    for key in ('kT', 'cross'):
        result[key] = np.array(result[key])
    return result


''' --------------------- CrossMapping --------------------- '''

class CrossMapping(SymmetricPairMapping):
    '''SymmetricPairMapping where values are CrossTable objects.

    smart: True, False, 'files', or 'defaults'
        if smart, values can also be strings.
        'files' --> strings are interepreted as filenames.
        'defaults' --> strings are interpreted as one of the keys for CrossTable.from_defaults.
        True --> use 'files' mode if the file exists, else 'defaults' mode, for strings.
        False --> strings are not allowed; all entries must be CrossTable objects.

        string values are converted into CrossTable objects immediately, during self[key] = value;
    '''
    cross_table_type = CrossTable

    def __init__(self, mapping=dict(), *, smart=True):
        self.smart = smart
        super().__init__(mapping)

    def __setitem__(self, key, value):
        '''set self.mapping[key] = value.
        key: tuple of (obj1, obj2). If obj2 < obj1, will use (obj2, obj1) instead.
        value: CrossTable object
        '''
        smart = self.smart
        if smart and isinstance(value, str):
            if isinstance(smart, str):
                if smart == 'files':
                    value = self.cross_table_type.from_file(value)
                elif smart == 'defaults':
                    value = self.cross_table_type.from_defaults(value)
                else:
                    raise ValueError(f'smart={smart!r} not recognized.')
            else:  # smart = True; try both modes.
                if os.path.isfile(value):
                    value = self.cross_table_type.from_file(value)
                else:
                    value = self.cross_table_type.from_defaults(value)
        else:  # not smart
            if not isinstance(value, CrossTable):
                raise TypeError(f'value={value!r} must be a CrossTable object')
        super().__setitem__(key, value)

    # # # DISPLAY # # #
    _repr_newlines = True  # put newlines bewteen each key-value pair in __repr__

    def __str__(self):
        keys = tuple(self.keys())
        vals = tuple(self.values())
        keys = [f'({key[0]!s}, {key[1]!s})' for key in keys]
        vals = [(str(val) if getattr(val, 'filebase', None) is None else val.filebase) for val in vals]
        contents = ', '.join(f'{key}: {val!r}' for key, val in zip(keys, vals))
        return f'{type(self).__name__}({contents})'



"""
File Purpose: Bifrost Stagger (align inputs to grid)

STAGGER ORDERS DEFINED HERE:
    1 - "simplest" method available.
            good enough, for most uses. ~20% faster than order=5
    5 - improved 5th order scheme.
            the improvement refers to improved precision for "shift" operations.

The implementations here use numpy.
Future implementations may use numba if numpy is too slow
    helita has numba implementation but numba can make installation more challenging;
    if adding numba here be sure to make it optional, not a requirement for PlasmaCalcs.

TESTED:
    seems to give same results as helita implementation, for 5th order numpy scheme.
    seems to take roughly the same amount of time as helita implementation
        (maybe 5 to 10% faster for shifts. Within 5% speed for derivatives.)


[TODO]
mod_J is not identical when comparing stagger_minimal_slicing=True & False modes,
    even though I would expect it to be identical in both cases.
    However, the result seems to be very close
        (e.g. J_z within 0.1% except for points with very small J)
    In the interest of time, I am saying "that's good enough" for now. (-SE, 2025/01/28)

Also, currently cannot slice using nontrivial step in multiple slices at once :(
"""

import collections
import os
import re

import numpy as np
import xarray as xr

from ...defaults import DEFAULTS
from ...dimensions import YZ_FROM_X
from ...errors import (
    InputError, InputConflictError, InputMissingError,
    DimensionalityError, DimensionSizeError,
    LoadingNotImplementedError,
)
from ...tools import (
    alias, alias_child, simple_property, weakref_property_simple,
    xarray_isel, xarray_copy_kw,
    is_iterable,
)
from ...quantities import QuantityLoader


''' --------------------- Stagger Constants --------------------- '''

XYZ_TO_INT = {'x': 0, 'y': 1, 'z': 2}
XYZ_TO_STR = {0: 'x', 1: 'y', 2: 'z', 'x': 'x', 'y': 'y', 'z': 'z'}
PAD_PERIODIC = 'wrap'     # how to pad periodic dimensions
PAD_NONPERIODIC = 'reflect'  # how to pad nonperiodic dimensions

StaggerConstants = collections.namedtuple('StaggerConstants', ('a', 'b', 'c'))

# # FIRST ORDER SCHEME # #
STAGGER_ABC_DERIV_o1 = StaggerConstants(1.0, 0, 0)
STAGGER_ABC_SHIFT_o1 = StaggerConstants(0.5, 0, 0)

# # FIFTH ORDER SCHEME # #
# derivatives
c = (-1 + (3**5 - 3) / (3**3 - 3)) / (5**5 - 5 - 5 * (3**5 - 3))
b = (-1 - 120*c) / 24
a = (1 - 3*b - 5*c)
STAGGER_ABC_DERIV_o5 = StaggerConstants(a, b, c)

# shifts (i.e. not a derivative)
c = 3.0 / 256.0
b = -25.0 / 256.0
a = 0.5 - b - c
STAGGER_ABC_SHIFT_o5 = StaggerConstants(a, b, c)

# remove temporary variables from the module namespace
del c, b, a


''' --------------------- Stagger Tools --------------------- '''

def transpose_to_0(array, x):
    '''move x (int) axis to the front of array (numpy array), by swapping with 0th axis.
    
    Note this is its own inverse, i.e. transpose_to_0(transpose_to_0(array, x), x) == array.
    '''
    tup = transpose_to_0_tuple(array.ndim, x)
    return np.transpose(array, tup)

def transpose_to_0_tuple(ndim, x):
    '''return tuple to pass to np.transpose to swap axis 0 with axis x (int).

    T = np.transpose(array, result) gives array with x axis swapped with 0th axis.
    np.transpose(T, result) swaps them back.
    '''
    if x < 0:
        x = x + ndim  # e.g. (ndim=3, x=-1) --> x=2
    if x >= ndim:
        raise DimensionalityError(f'axis x={x} is out of bounds for ndim={ndim}')
    result = list(range(ndim))
    result[x] = 0
    result[0] = x
    return tuple(result)

def simple_slice(start_shift, end_shift):
    '''return slice(start_shift, end_shift), but if end_shift is 0 use None instead.'''
    return slice(start_shift, None if end_shift == 0 else end_shift)


class StaggerPrePadManager3D():
    '''manages pre-padding for stagger operations in 'x', 'y', 'z' dimensions.
    (3D stagger operations' inputs and outputs are always 3D arrays.)

    ops: str
        string of operations to do, separated by whitespace or '_'.
        each operation must be one of:
            'xup',   'xdn',   'yup',   'ydn',   'zup',   'zdn',
            'ddxup', 'ddxdn', 'ddyup', 'ddydn', 'ddzup', 'ddzdn'.
    slices: dict of indexers or None
        dict of 'x', 'y', 'z' to indexers (e.g. from a QuantityLoader).
        None --> empty dict.
    size: dict of int
        lengths of 'x', 'y', 'z' dimensions.

    calling self returns padded_slices:
        dict of 'x', 'y', 'z' slices to use to pad an array before doing this series of ops,
        such that the result will be properly aligned & shaped the same as non-staggered vars.
        Note: dims without slice will have result[dim] = None.
        dims with slice will have result[dim] =
            sum of self.PAD_AMOUNT['up' or 'dn'] for all ops with dim.
            E.g. 'xup ddxup xdn' --> result['x'] = (2,3) + (2,3) + (3,2) = (7,8).
    '''
    PAD_AMOUNT = {'up': (2,3), 'dn': (3,2)}

    def __init__(self, ops, slices=None, *, sizes):
        self.ops = ops.replace('_', ' ').split()
        self.slices = {} if slices is None else slices
        self.sizes = sizes
        self.pad_amounts = self.get_pad_amounts()
        self.modes = {}  # padding modes determined during self.get_prepad_slices(). Helps with debugging.
        self.to_squeeze_after = set()   # dims to squeeze after stagger operations are finished.

    def get_pad_amounts(self):
        '''return amounts of padding for each dim, based on ops (ignoring self.slices for now).'''
        padding = {x: (0,0) for x in ('x', 'y', 'z')}
        for op in self.ops:
            dd, x, up = re.fullmatch(r'(dd)?([xyz])(up|dn)', op).groups()
            pad_x = padding[x]
            if pad_x is not None:
                amount = self.PAD_AMOUNT[up]
                padding[x] = (pad_x[0] + amount[0],
                              pad_x[1] + amount[1])
        return padding

    def cumulative_pad_amounts(self):
        '''return amounts of padding for each dim, after each op.
        e.g. [{'x':(0,0),'y':(0,0),'z':(2,3)}, {'x':(2,3),'y':(0,0),'z':(2,3)}, ...]
        '''
        result = []
        padding = {x: (0,0) for x in ('x', 'y', 'z')}
        for op in self.ops:
            dd, x, up = re.fullmatch(r'(dd)?([xyz])(up|dn)', op).groups()
            pad_x = padding[x]
            if pad_x is not None:
                amount = self.PAD_AMOUNT[up]
                padding[x] = (pad_x[0] + amount[0],
                              pad_x[1] + amount[1])
            result.append(padding.copy())
        return result

    def get_prepad_slice(self, x):
        '''returns padded slice (or other indexer) for this dim ('x', 'y', or 'z')
        None if self.pad_amount[x] is None.
        None if self.slices[x] == slice(None).
        self.slices[x] if self.pad_amount[x] is (0,0).
        '''
        indexer = self.slices.get(x, None)
        if indexer is None:
            self.modes[x] = None
            return None
        if isinstance(indexer, slice) and indexer == slice(None):
            self.modes[x] = None
            return None
        pad_amount = self.pad_amounts[x]
        if pad_amount == (0,0):  # no ops occuring along this dim --> use original indexer.
            self.modes[x] = None
            if (not isinstance(indexer, slice)) and (not is_iterable(indexer)):  # i.e., indexer is int.
                # int indexers need to be converted to slice before stagger (can squeeze later),
                #  because stagger inputs and outputs are always 3D arrays.
                self.to_squeeze_after.add(x)  # need to squeeze this dim after stagger ops are completed.
                return slice(indexer, indexer+1)
            else:
                return indexer   # (still use original indexer though!)
        if is_iterable(indexer):
            errmsg = (f'{type(self).__name__}.get_prepad_slice(x={x!r}) with iterable indexer.\n'
                      f'Expected None, slice, or int; got {indexer}.\n'
                      '(This might be avoided if stagger_minimal_slicing=False)')
            raise NotImplementedError(errmsg)
        pad_start, pad_stop = pad_amount
        if isinstance(indexer, slice):
            start = indexer.start
            stop = indexer.stop
            step = indexer.step
            if step is None: step = 1
            if step == 1:
                self.modes[x] = 'simple_slice'
            else:
                self.modes[x] = 'step_slice'
        else:   # assume int.
            start = indexer
            stop = indexer + 1
            step = 1
            self.modes[x] = 'simple_int'
            self.to_squeeze_after.add(x)  # need to squeeze this dim after stagger ops are completed.
        # handle start & stop -- cleanup
        if x not in self.sizes:
            raise InputError(f'provided indexer for {x!r} but {x!r} not provided in self.sizes.')
        L = self.sizes[x]
        if start is None:
            start = 0
        if stop is None:
            stop = L
        if start < 0:
            start = L + start
            if start < 0:
                start = 0
        if stop < 0:
            stop = L + stop
            if stop < 0:
                stop = 0
        if start < pad_start:  # not enough room to pad easily on the left.
            errmsg = (f'{type(self).__name__}.get_prepad_slice(x={x!r}) '
                      f'when start (={start}) < pad_start (={pad_start}).\n'
                      '(This might be avoided if stagger_minimal_slicing=False)')
            raise NotImplementedError(errmsg)
        if stop > L - pad_stop:  # not enough room to pad easily on the right.
            errmsg = (f'{type(self).__name__}.get_prepad_slice(x={x!r}) '
                      f'when stop (={stop}) > len-pad_stop (={L-pad_stop}).\n'
                      '(This might be avoided if stagger_minimal_slicing=False)')
            raise NotImplementedError(errmsg)
        # handle step
        if step is None or step == 1:   # trivial step -- easy!
            start = start - pad_start
            stop = stop + pad_stop
            return slice(start, stop)
        elif step <= pad_start + pad_stop:
            errmsg = (f'{type(self).__name__}.get_prepad_slice(x={x!r}) '
                      f'when step ({step}) != 1 (or None), and step <= pad_start + pad_stop ({pad_start}+{pad_stop}).\n'
                      '(This might be avoided if stagger_minimal_slicing=False)')
            raise NotImplementedError(errmsg)
        else:   # create multiple independent regions.
            # E.g. slice(20, 41, 10), pad (2,3) --> [18,19,20,21,22,23, 28,29,30,31,32,33, 38,39,40,41,42,43]
            #  (note, slice(20, 40, 10) will exclude the 38-43 because unpadded would exclude the point at 40.)
            # for now, just use a python for loop for better readability. [TODO][EFF] if slow, improve efficiency.
            rr = range(L)
            result = []
            here = start
            while here < stop:
                region = list(rr[here - pad_start : here + pad_stop + 1])  # +1 so that here+pad_stop is included too.
                result = result + region
                here = here + step
            return result
            # note, using multiple independent regions is not particularly efficient;
            #   running it though the Staggerer will end up computing lots of meaningless values,
            #   e.g. in the example above, the result will have the values at:
            #       [20,21,22,23, 28,29,30,31,32,33, 38,39,40],
            #   and stagger_postprocess will return only the values at [20, 30, 40],
            #       since those are the values we care about.
            #   (also note, the value at 20 will be from [18,19,20,21,22,23] appropriately,
            #       however the value at 22, e.g., will be from [20,21,22,23,28,29],
            #       which isn't even the correct value at that point.)
            # however, it's definitely more efficient to do this, than to stagger the ENTIRE array,
            #   especially when step is large.
        assert False, 'coding error if reached this line.'

    __call__ = alias('get_prepad_slices')

    def get_prepad_slices(self):
        '''return padded_slices to use.
        like self.slices but with padding added to provided indexers.
        see help(type(self)) for more details.
        '''
        self.modes = {}  # <-- reset self.modes before getting all the prepad slices.
        self.to_squeeze_after = set()   # <-- reset self.to_squeeze_after, too.
        padded_slices = {}
        for x in ('x', 'y', 'z'):
            padded_x = self.get_prepad_slice(x)
            if padded_x is not None:
                padded_slices[x] = padded_x
        self._prepad_slices_result = padded_slices
        return padded_slices

    def get_prepad_dims(self):
        '''return dict of which dims are adjusted by prepadding.'''
        self.get_prepad_slices()
        modes = self.modes
        return list(x for x, mode in modes.items() if mode is not None)

    def _step_slice_unpadder_in_postprocess(self, x):
        '''unpadder for step_slice, in postprocessing.'''
        padx = self.pad_amounts[x]
        stepx = padx[0] + padx[1] + 1
        return slice(None, None, stepx)

    def stagger_postprocess(self, array):
        '''postprocessing of array after stagger ops, to deal with padding details.
        Here, does the following:
            (1) ensure that array is 3D. (probably a numpy array, but xarray is permitted too.)
            (2) if any modes are 'step_slice', slice to recover only the relevant values.
            (3) squeeze array after stagger ops, to respect any (unpadded) int indexers.
                Also asserts that dims to squeeze actually have size 1.
        '''
        # (1)
        if array.ndim != 3:
            raise DimensionalityError(f'expected 3D array; got {array.ndim}D.')
        # (2)
        if any(mode == 'step_slice' for mode in self.modes.values()):
            slices = [slice(None), slice(None), slice(None)]
            for i, x in enumerate(('x', 'y', 'z')):
                if self.modes.get(x, None) == 'step_slice':
                    slices[i] = self._step_slice_unpadder_in_postprocess(x)
            array = array[tuple(slices)]
        # (3)
        if 'z' in self.to_squeeze_after:
            if array.shape[2] != 1:
                raise DimensionSizeError(f'expected size 1 along z; got {array.shape[2]}.')
            array = array[:, :, 0]
        if 'y' in self.to_squeeze_after:
            if array.shape[1] != 1:
                raise DimensionSizeError(f'expected size 1 along y; got {array.shape[1]}.')
            array = array[:, 0]
        if 'x' in self.to_squeeze_after:
            if array.shape[0] != 1:
                raise DimensionSizeError(f'expected size 1 along x; got {array.shape[0]}.')
            array = array[0]
        return array

    def unpad1D(self, x, array):
        '''remove padding from 1D array corresponding to x dimension.
        assumes this is being called after self.get_prepad_slices()
            (i.e., expects self.modes and self.to_squeeze_after, to be set appropriately.)
        x: 'x', 'y', or 'z'.

        if self.to_squeeze_after[x], result will be a scalar instead of a 1D array.
        '''
        if array.ndim != 1:
            raise DimensionalityError(f'expected 1D array; got {array.ndim}D.')
        if x in self._prepad_slices_result:
            pad = self.pad_amounts[x]
            mode = self.modes[x]
            if mode is None:
                result = array
            elif mode == 'step_slice':
                result = array[pad[0] : -pad[1]]
                result = result[self._step_slice_unpadder_in_postprocess(x)]
            else:
                result = array if pad==(0,0) else array[pad[0] : -pad[1]]
        else:
            result = array
        if x in self.to_squeeze_after:
            if result.size != 1:
                raise DimensionSizeError(f'expected size 1; got {result.size}.')
            result = result[0]
        return result

    def xarray_copy_kw_from_unstaggered(self, unstaggered):
        '''return xarray_copy_kw(unstaggered), updated to remove padding.
        E.g.:
            staggered = stagger(unstaggered, op, pre_padded=self.get_prepad_slices())
            kw_data_array = self.xarray_copy_kw_from_unstaggered(unstaggered)
            as_data_array = xr.DataArray(staggered, **kw_data_array)
        '''
        result = xarray_copy_kw(unstaggered)
        coords = result['coords']
        dims = result['dims']
        for x in ('x', 'y', 'z'):
            if x in coords:
                coords[x] = self.unpad1D(x, coords[x])  # unpad1D also handles squeezing if needed.
        if any(x in dims for x in self.to_squeeze_after):
            result['dims'] = (x for x in dims if x not in self.to_squeeze_after)
        return result

    def op_prepad_kws(self):
        '''return list of dicts to use for stagger.do(array, op, **kw), for op in self.ops.
        [TODO] handle left_first=False.
        '''
        slices = self.get_prepad_slices()
        modes = self.modes
        padops = self.cumulative_pad_amounts()
        result = []
        for op, padop in zip(self.ops, padops):
            opresult = dict()
            dd, x, up = re.fullmatch(r'(dd)?([xyz])(up|dn)', op).groups()
            if modes.get(x, None) is not None:
                opresult['pre_padded'] = slices[x]
            if dd:
                opresult['depad_dx'] = padop[x]
            result.append(opresult)
        return result

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'ops={self.ops!r}',
                    f'slices={self.slices!r}',
                    f'sizes={self.sizes!r}',
                    f'pad_amounts={self.pad_amounts!r}',
                    f'modes={self.modes!r}'
                    ]
        return f'{type(self).__name__}({", ".join(contents)})'


''' --------------------- Staggerer --------------------- '''

class Staggerer():
    '''class to do staggering along an axis.
    internally, staggering will transpose, stagger along 0th axis, then transpose back.
    
    x: int or str
        the axis to take the derivative along.
        str --> 'x', 'y', or 'z', corresponding to 0, 1, 2.
        internally stored as int.
    periodic: bool
        whether to treat the array as periodic along x.
        True --> use pad='wrap' to fill values for stagger at edges of array.
        False --> use pad='reflect' to fill values for stagger at edges of array.
    dx: None, number, or 1D array
        spacing along this axis. relevant for deriv but not shift.
        None --> deriv methods will crash.
    order: 1 or 5.
        order of the scheme to use, by default.
    mode: str
        method for stagger calculations. Right now, only supports 'numpy_improved'.
        Eventually might support 'numpy', 'numba', and 'numba_improved'.
    short_ok: bool
        whether it is okay for arrays to be too short along x axis.
        True --> if too short: for shift, return as-is; for derivative, return zeros.
        False --> if too short: raise DimensionSizeError
    assert_ndim: None or int
        if provided, assert ndim(array) equals this value, before staggering.
        e.g. if expecting all arrays to be 3D, use assert_ndim=3.
    '''
    PAD_AMOUNT = {'up': (2,3), 'dn': (3,2)}

    def __init__(self, x, *, periodic, dx=None, order=5, mode='numpy_improved',
                 short_ok=True, assert_ndim=None):
        self.x = XYZ_TO_INT.get(x, x)   # self.x is an int.
        self.xstr = XYZ_TO_STR[self.x]
        self.periodic = periodic
        self.order = order
        self.dx = dx
        self.mode = mode
        if mode != 'numpy_improved':
            raise NotImplementedError(f'mode={mode!r}. Only implemented mode="numpy_improved" so far.')
        self.short_ok = short_ok
        self.assert_ndim = assert_ndim

    # # # PROPERTIES # # #
    dx = simple_property('_dx',
            doc='''spacing along this axis. relevant for deriv but not shift.
            if None, getting self.dx will raise InputMissingError.''')
    @dx.getter
    def dx(self):
        result = getattr(self, '_dx', None)
        if result is None:
            raise InputMissingError('dx is None, but dx required for derivatives.')
        return result

    # # # SIZE CHECK # # #
    def size_x(self, array):
        '''returns size of array along x axis.'''
        return array.shape[self.x]

    def at_least_size_x(self, array, size, *, short_ok=None):
        '''returns whether array has at least this size along x axis.
        short_ok: None or bool
            whether it is okay for arrays to be too short along x axis.
            None --> use self.short_ok.
            False --> raise DimensionSizeError if too short.
        '''
        if short_ok is None:
            short_ok = self.short_ok
        size_x = self.size_x(array)
        if size_x < size:
            if short_ok:
                return False
            else:
                errmsg = f'array too short in axis {self.x}: got {size_x}; expected >= {size}.'
                raise DimensionSizeError(errmsg)
        else:
            return True

    # # # STAGGER PREP & POST # # #
    def pad(self, transposed_array, *, up):
        '''pad array along 0th axis as appropriate, to prepare it for staggering computations.
        up=True (up) or False (down)
        '''
        pad_mode = PAD_PERIODIC if self.periodic else PAD_NONPERIODIC
        pad_amount = self.PAD_AMOUNT['up' if up else 'dn']
        np_padding = [pad_amount] + [(0, 0)] * (transposed_array.ndim - 1)
        return np.pad(transposed_array, np_padding, mode=pad_mode)

    def transpose(self, array):
        '''swap axis self.x with axis 0.
        Note, this is its own inverse, i.e. transpose(transpose(array)) == array.
        '''
        return transpose_to_0(array, self.x)

    def _pre_stagger_prep(self, array, *, up, pre_padded=False):
        '''return array prepped for staggering.
        i.e., transpose to put x axis at front, then pad appropriately.
        Also ensure array ndim == self.assert_ndim if provided.

        pre_padded: None, bool, iterable, or slice
            bool(pre_padded) tells whether array has already been padded in the relevant dim.
        '''
        if self.assert_ndim is not None:
            if array.ndim != self.assert_ndim:
                errmsg = f'array ndim ({array.ndim}) != assert_ndim ({self.assert_ndim})'
                raise DimensionalityError(errmsg)
        array_transpose = self.transpose(array)
        if is_iterable(pre_padded) or bool(pre_padded):   # is_iterable checked to avoid bool(numpy array)
            return array_transpose
        else:
            return self.pad(array_transpose, up=up)

    def _post_stagger(self, array, **kw_None):
        '''return array converted back to original format, post-staggering.
        i.e., transpose to put x axis from front back to original position.
        '''
        return self.transpose(array)

    def _stagger_dx_padding(self, dx, *, up, pre_padded=False, depad_dx=None):
        '''returns dx, or dx[pre_padded][unpad] if pre_padded provided.
        unpad = depad_dx if provided, else slice to remove padding implied by self & up.
        '''
        if is_iterable(dx) and (is_iterable(pre_padded) or bool(pre_padded)):
            result = dx[pre_padded]
            if depad_dx is None:
                up = 'up' if up else 'dn'
                result = result[self.PAD_AMOUNT[up][0] : -self.PAD_AMOUNT[up][1]]
            else:
                result = result[depad_dx[0] : -depad_dx[1]]
            return result
        else:
            return dx

    # # # SHIFTS # # #
    def shift(self, array, *, up, **kw):
        '''shift array along x axis, staggering as appropriate.
        up=True (up) or False (down)
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        order = self.order
        if order == 1:
            return self.shift_o1(array, up=up, **kw)
        elif order == 5:
            return self.shift_o5(array, up=up, **kw)
        else:
            raise InputError(f'order={order} not supported; must be 1 or 5')

    def shift_o1(self, array, *, up, pre_padded=False, **kw_None):
        '''first order shift, staggering up.'''
        raise NotImplementedError('[TODO]')  # could use 

    def shift_o5(self, array, *, up, pre_padded=False, **kw_None):
        '''fifth order shift, staggering as appropriate.'''
        if not self.at_least_size_x(array, 5):
            return array
        f = self._pre_stagger_prep(array, up=up, pre_padded=pre_padded)
        A, B, C = STAGGER_ABC_SHIFT_o5
        upshift = 1 if up else 0
        Pstart, Pend = self.PAD_AMOUNT['up' if up else 'dn']
        start = Pstart + upshift
        end = -Pend + upshift
        i00 = simple_slice(start + 0, end + 0)
        i1p = simple_slice(start + 1, end + 1)
        i2p = simple_slice(start + 2, end + 2)
        i1m = simple_slice(start - 1, end - 1)
        i2m = simple_slice(start - 2, end - 2)
        i3m = simple_slice(start - 3, end - 3)
        f0 = f[i00]
        out = f0 + (A * (            f[i1m]-f0) +
                    B * (f[i1p]-f0 + f[i2m]-f0) +
                    C * (f[i2p]-f0 + f[i3m]-f0))
        result = self._post_stagger(out)
        return result

    # # # DERIVATIVES # # #
    def deriv(self, array, *, up, **kw):
        '''take derivative of array along x axis, staggering as appropriate.
        up=True (up) or False (down)
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        order = self.order
        if order == 1:
            return self.deriv_o1(array, up=up, **kw)
        elif order == 5:
            return self.deriv_o5(array, up=up, **kw)
        else:
            raise InputError(f'order={order} not supported; must be 1 or 5')

    def deriv_o1(self, array, *, up, pre_padded=False, depad_dx=None):
        '''first order derivative, staggering up.'''
        raise NotImplementedError('[TODO]')

    def deriv_o5(self, array, *, up, pre_padded=False, depad_dx=None):
        '''fifth order derivative, staggering as appropriate.'''
        if not self.at_least_size_x(array, 5):
            return np.zeros_like(array)
        dx = self.dx
        if 0 < np.ndim(dx) < np.ndim(array):  # add enough np.newaxis for dx dims if needed.
            dx = np.expand_dims(dx, tuple(range(np.ndim(dx), np.ndim(array))))
        dx = self._stagger_dx_padding(dx, up=up, pre_padded=pre_padded, depad_dx=depad_dx)
        f = self._pre_stagger_prep(array, up=up, pre_padded=pre_padded)
        A, B, C = STAGGER_ABC_DERIV_o5
        upshift = 1 if up else 0
        Pstart, Pend = self.PAD_AMOUNT['up' if up else 'dn']
        start = Pstart + upshift
        end = -Pend + upshift
        i00 = simple_slice(start + 0, end + 0)
        i1p = simple_slice(start + 1, end + 1)
        i2p = simple_slice(start + 2, end + 2)
        i1m = simple_slice(start - 1, end - 1)
        i2m = simple_slice(start - 2, end - 2)
        i3m = simple_slice(start - 3, end - 3)
        out = dx * (A * (f[i00] - f[i1m]) + 
                    B * (f[i1p] - f[i2m]) +
                    C * (f[i2p] - f[i3m]))
        result = self._post_stagger(out)
        return result

    # # # SHORTHANDS # # #
    def shift_up(self, array, **kw):
        '''shift array up. Equivalent: shift(array, up=True)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.shift(array, up=True, **kw)

    def shift_dn(self, array, **kw):
        '''shift array down. Equivalent: shift(array, up=False)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.shift(array, up=False, **kw)

    def deriv_up(self, array, **kw):
        '''take derivative of array up. Equivalent: deriv(array, up=True)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.deriv(array, up=True, **kw)

    def deriv_dn(self, array, **kw):
        '''take derivative of array down. Equivalent: deriv(array, up=False)'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.deriv(array, up=False, **kw)

    up = alias('shift_up')
    dn = alias('shift_dn')
    ddup = alias('deriv_up')
    dddn = alias('deriv_dn')

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'x={self.x!r}',
                    f'periodic={self.periodic}',
                    f'order={self.order}'
                    ]
        if self._dx is None:
            contents.append('dx=None')
        if self.mode != 'numpy_improved':
            contents.append(f'mode={self.mode!r}')
        return f'{type(self).__name__}({", ".join(contents)})'


''' --------------------- Stagger Interface 3D --------------------- '''

class StaggerInterface3D():
    '''class to do staggering along 'x', 'y', or 'z', axes.
    Call self(array, opstr) to do the operation(s) implied by opstr (from left to right)
        E.g. self(array, 'xup ddzdn') --> shift array up x, then take z derivative down.

    periodic_x, periodix_y, periodic_z: bool
        whether to treat (unsliced) arrays as periodic along x, y, z axes.
        True --> use pad='wrap' to fill values for stagger at edges of array.
        False --> use pad='reflect' to fill values for stagger at edges of array.
    dx, dy, dz: None, number, or 1D array
        spacing along each axis. relevant for deriv but not shift.
        None --> deriv methods will crash.
    order: 1 or 5.
        order of the scheme to use, by default.
    mode: str
        method for stagger calculations. Right now, only supports 'numpy_improved'.
        Eventually might support 'numpy', 'numba', and 'numba_improved'.

    self.x, self.y, self.z store Staggerer objects for each axis.
    '''
    staggerer_cls = Staggerer
    PAD_AMOUNT = {'up': (2,3), 'dn': (3,2)}

    def __init__(self, *,
                 periodic_x, periodic_y, periodic_z,
                 dx=None, dy=None, dz=None,
                 order=5, mode='numpy_improved'):
        kw_shared = dict(order=order, mode=mode, assert_ndim=3)
        self.x = self.staggerer_cls('x', periodic=periodic_x, dx=dx, **kw_shared)
        self.y = self.staggerer_cls('y', periodic=periodic_y, dx=dy, **kw_shared)
        self.z = self.staggerer_cls('z', periodic=periodic_z, dx=dz, **kw_shared)

    # # # ALIASES - ORDER AND MODE # # #
    order = alias_child('x', 'order')
    @order.setter
    def order(self, value):
        self.x.order = value
        self.y.order = value
        self.z.order = value

    mode = alias_child('x', 'mode')
    @mode.setter
    def mode(self, value):
        self.x.mode = value
        self.y.mode = value
        self.z.mode = value

    # # # ALIASES - OPERATIONS # # #
    xup = alias_child('x', 'shift_up')
    xdn = alias_child('x', 'shift_dn')
    yup = alias_child('y', 'shift_up')
    ydn = alias_child('y', 'shift_dn')
    zup = alias_child('z', 'shift_up')
    zdn = alias_child('z', 'shift_dn')

    ddxup = alias_child('x', 'deriv_up')
    ddxdn = alias_child('x', 'deriv_dn')
    ddyup = alias_child('y', 'deriv_up')
    ddydn = alias_child('y', 'deriv_dn')
    ddzup = alias_child('z', 'deriv_up')
    ddzdn = alias_child('z', 'deriv_dn')

    # # # DO / CALL # # #
    def do(self, array, opstr, *, prepadder=None, left_first=True):
        '''do the operation(s) implied by opstr.

        opstr: str
            string of operations to do, separated by whitespace or '_'.
            each operation must be one of:
                'xup',   'xdn',   'yup',   'ydn',   'zup',   'zdn',
                'ddxup', 'ddxdn', 'ddyup', 'ddydn', 'ddzup', 'ddzdn'.
        prepadder: None or StaggerPrePadManager3D
            manages prepadding info for array.
            None --> array has not been prepadded; need to pad internally.
            else --> array has already been prepadded.
        left_first: bool
            when multiple operations, tells order in which to apply them.
            True --> apply operations from left to right.
                    E.g. 'xup ddzdn' --> first shift x up, then take z derivative down.
            False --> apply operations from right to left.
                    E.g. 'xup ddzdn' --> first take z derivative down, then shift x up.
        '''
        ops = opstr.replace('_', ' ').split()
        if not left_first:
            ops = ops[::-1]
        if prepadder is None:
            op_prepad_kws = [{} for op in ops]
        else:
            op_prepad_kws = prepadder.op_prepad_kws()
        for op, kw in zip(ops, op_prepad_kws):
            x = op.lstrip('dd')[0]  # 'x', 'y', or 'z'
            xup = getattr(self, op)  # one of the ops, e.g. self.xup or self.ddzdn
            array = xup(array, **kw)
        return array

    __call__ = alias('do')

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'periodic_x={self.x.periodic}',
                    f'periodic_y={self.y.periodic}',
                    f'periodic_z={self.z.periodic}',
                    f'order={self.order}'
                    ]
        if self.x.dx is None:
            contents.append('dx=None')
        if self.y.dx is None:
            contents.append('dy=None')
        if self.z.dx is None:
            contents.append('dz=None')
        if self.mode != 'numpy_improved':
            contents.append(f'mode={self.x.mode!r}')
        return f'{type(self).__name__}({", ".join(contents)})'


''' --------------------- Stagger Interface 3D Haver --------------------- '''

class BifrostStaggerable(QuantityLoader):
    '''manages stagger stuff for BifrostCalculator'''
    stagger_interface_cls = StaggerInterface3D
    stagger_prepad_cls = StaggerPrePadManager3D

    stagger = simple_property('_stagger', setdefaultvia='_create_stagger_interface',
            doc='''stagger interface object, for doing staggering operations.
            BifrostCalculator staggers values to grid cell centers upon loading, by default.
            When all values are at cell centers, can proceed without any more staggering.
            Note: this object assumes lengths are in 'raw' units when doing derivatives.''')

    def _create_stagger_interface(self):
        '''create stagger interface based on self.params'''
        params = self.params
        kw = {}
        kw['periodic_x'] = params['periodic_x']
        kw['periodic_y'] = params['periodic_y']
        kw['periodic_z'] = params['periodic_z']
        if self.has_meshfile():
            mesh_coords = self.load_mesh_coords()
            kw['dx'] = mesh_coords['dx']
            kw['dy'] = mesh_coords['dy']
            kw['dz'] = mesh_coords['dz']
        else:
            if 'meshfile' in self.params:  # but not self.has_meshfile()
                print('warning: meshfile file does not exist',
                      f'(but meshfile param is defined in .idl file): {self.meshfile!r}.',
                      'Defaulting to dx, dy, dz from .idl file, during stagger operations.')
            kw['dx'] = params['dx']
            kw['dy'] = params['dy']
            kw['dz'] = params['dz']  # might crash... but that's ok;
            # PlasmaCalcs doesn't yet know how to handle different dz for each snapshot.
        return self.stagger_interface_cls(**kw)

    # # # BEHAVIOR ATTRS # # #
    cls_behavior_attrs.register('stagger_order', default=5)
    stagger_order = alias_child('stagger', 'order')
    cls_behavior_attrs.register('stagger_mode', default='numpy_improved')
    stagger_mode = alias_child('stagger', 'mode')

    # # # PADDING / MINIMAL SLICES # # #
    cls_behavior_attrs.register('stagger_minimal_slicing', default=False)
    stagger_minimal_slicing = simple_property('_stagger_minimal_slicing',
        setdefaultvia='_default_stagger_minimal_slicing',
        doc='''whether to use "smart" minimal slicing when getting stagger vars.
        default depends on self.maindims_full_size and DEFAULTS.ARRAY_MBYTES_MAX
            (default = True if maindims arrays smaller than 1/3rd max size, else False.)
            (eventually, might always use default=True, but currently lacking some essential features,
                like slice with step!=1. The default here disables those crashes for small enough snaps,
                but enables the far-superior efficiency of minimal slicing, for very large snaps.)

        When True, e.g. self('xup_var', slices=dict(x=slice(10, 20))
            would get prepadder for  array = self('var', slices=dict(x=slice(8, 23)),
            then self.stagger(array, 'xup', prepadder=dict(x=slice(8, 23)).
        CAUTION: calling self.stagger.do directly will not incorporate minimal slicing.
        Note: the result should be equivalent to when not using minimal slicing,
            [EFF] but using minimal slicing is much more efficient for large arrays.''')

    def _default_stagger_minimal_slicing(self):
        '''returns default value to use for stagger_minimal_slicing.
        The default is True if maindims_full_size implies arrays bigger than ARRAY_MBYTES_MAX/3.
        See help(type(self).stagger_minimal_slicing) for more details.
        '''
        size = self.maindims_full_size
        check_dtype = DEFAULTS.ARRAY_MEMORY_CHECK_DTYPE_MIN
        maindims_mbytes = size * np.dtype(check_dtype).itemsize / 1024**2
        return (maindims_mbytes > DEFAULTS.ARRAY_MBYTES_MAX / 3)

    def stagger_prepad_manager(self, ops):
        '''get stagger prepad manager to use for padding an array before doing this series of ops,
        such that the result will be properly aligned & shaped the same as non-staggered vars.
        '''
        slices = self.standardized_slices()
        sizes = self.maindims_full_sizes
        return self.stagger_prepad_cls(ops, slices=slices, sizes=sizes)

    def stagger_prepad_slices(self, ops):
        '''get self.slices to use for padding an array before doing this series of ops,
        such that the result will be properly aligned & shaped the same as non-staggered vars.
        '''
        return self.stagger_prepad_manager(ops).get_prepad_slices()

    def _stagger_direct_and_slice(self, array, ops):
        '''return result of applying these stagger ops to array, and slicing by self.slices.
        
        array: numpy array with same shape as self.maindims_full_shape
            e.g. result of np.memmap, inside the self.load_fromfile method.
        ops: str
            ops to do, separated by whitespace or '_'. Passed to stagger methods.
        '''
        if self.stagger_minimal_slicing:
            if (not self._slice_maindims_in_load_direct):
                errmsg = 'stagger_minimal_slicing=True but _slice_maindims_in_load_direct=False'
                raise LoadingNotImplementedError(errmsg)
            prepadder = self.stagger_prepad_manager(ops)
            prepad_slices = prepadder.get_prepad_slices()
            with self.using(slices=prepad_slices):
                array = self._slice_direct_numpy(array)
            result = self.stagger(array, ops, prepadder=prepadder)
            result = prepadder.stagger_postprocess(result)
        else:
            result = self.stagger(array, ops)
            result = self._slice_direct_numpy(result)
        return result

    def _staggering_minimal_slicing_with_nontrivial_step(self, ops_or_axes):
        '''returns whether self.stagger_minimal_slicing and self.slices have nontrivial step relevant to ops.
        ops_or_axes: str
            ops, separated by '_' or whitespace.
            or, string containing all relevant axes, e.g. 'xyz'.
        '''
        if not self.stagger_minimal_slicing:
            return False
        if not self.slices:
            return False
        for x in ('x', 'y', 'z'):
            if x in ops_or_axes:
                slicex = self.slices.get(x, None)
                if isinstance(slicex, slice):
                    if (slicex.step is not None) and slice.step != 1:
                        return True
                elif is_iterable(slicex):
                    return True
        return False

    # # # VARS # # #
    @known_pattern(r'((?:(?:dd)?[xyz](?:up|dn)_)+)(.+)', deps=[1])  # e.g. xup_r, ddzdn_r, yup_ddydn_n
    def get_stagger_var(self, var, *, _match=None):
        '''stagger var, shifting and/or taking derivatives, up or down, by half grid cell.
        E.g. xup_r --> shift r up along x axis.
        E.g. ddxup_r --> take x derivative of r up along x axis.

        Multiple ops can be combined, e.g. xup_ddxdn_r
            --> first shift r up along x, then take x derivative down.
        '''
        # implementation notes:
        # (?:) is syntax for a "non-capturing group" in a regular expression;
        #    the match.groups()[0] will contain a string like op1_op2_op3_...opN_.
        # grouped these together into one pattern instead of letting QuantityLoader split it up,
        #    so that chained stagger ops only require padding once, if using stagger_minimal_slicing.
        #    this is usually just for efficiency, but essential if slices have non-trivial step (!=1);
        #    because the first op would pad with iterable of indices in that case,
        #       (e.g. [18,19,20,21,22,23, 28,29,30,31,32,33, ...]),
        #    and currently there's no implementation for further padding a list like that.
        #    But, grouping into one pattern means the padding will just be larger, but only done once.

        # preprocessing bookkeeping
        ops, var = _match.groups()
        ops.rstrip('_')  # remove trailing '_'. Now ops is like, e.g., 'xup' or 'xup_ddxdn_zup_ydn' ...
        if self.stagger_minimal_slicing:
            prepadder = self.stagger_prepad_manager(ops)
            slices = prepadder.get_prepad_slices()
        else:
            prepadder = None
            slices = None
        # >> actual computations <<
        unstaggered = self(var, slices=slices, squeeze_direct=False)
        result = self.stagger(unstaggered.values, ops, prepadder=prepadder)
        # postprocessing bookkeeping
        if 'dd' in ops:  # stagger assumes raw length for dx; need to convert units.
            # divides by u('l') for each derivative op (contains 'dd') in ops.
            result = result / self.u('length', convert_from='raw')**(ops.count('dd'))
        if prepadder is None:
            copy_kw = xarray_copy_kw(unstaggered)
        else:
            copy_kw = prepadder.xarray_copy_kw_from_unstaggered(unstaggered)
            result = prepadder.stagger_postprocess(result)
        result = xr.DataArray(result, **copy_kw)
        if self.slices and not self.stagger_minimal_slicing:
            result = xarray_isel(result, self.slices)
        if self.squeeze_direct:
            result = self._squeeze_direct_result__post_slicing(result)
        return result

    @known_pattern(r'facecurl_(.+)', deps=[0])
    def get_facecurl(self, var, *, _match=None):
        '''return curl of face-centered var, staggered from cell faces to cell edges
        E.g. facecurl_B --> curl of B, staggered to cell edges:
            (ddydn_B_z - ddzdn_B_y, ddzdn_B_x - ddxdn_B_z, ddxdn_B_y - ddydn_B_x).

        Assumes, but does not check, that var is face-centered, e.g. B, u, xup_r.

        [EFF] can make more efficient, if slow; current implementation gets component twice.
        '''
        here, = _match.groups()
        curls = []
        for x in self.iter_component():
            y, z = YZ_FROM_X[x]
            ddy_vz = self(f'dd{y}dn_{here}_{z}')
            ddz_vy = self(f'dd{z}dn_{here}_{y}')
            curl_x = self.assign_component_coord(ddy_vz - ddz_vy, x)
            curls.append(curl_x)
        result = self.join_components(curls)
        return result

    @known_pattern(r'centered_facecurl_(.+)', deps=[0])
    def get_centered_facecurl(self, var, *, _match=None):
        '''return curl of face-centered var, staggered from cell faces to cell centers
        E.g. centered_facecurl_B --> curl of B, staggered to cell centers:
            at_edges = (ddydn_B_z - ddzdn_B_y, ddzdn_B_x - ddxdn_B_z, ddxdn_B_y - ddydn_B_x).
            result = (yup_zup(at_edges_x), zup_xup(at_edges_y), xup_yup(at_edges_z)).

        Assumes, but does not check, that var is face-centered, e.g. B, u, xup_r.

        [EFF] can make more efficient, if slow; current implementation gets component twice.
        '''
        here, = _match.groups()
        curls = []
        simple_mode = True
        if self.stagger_minimal_slicing:
            if self.current_n_component() == 1:
                y, z = YZ_FROM_X[self.component_list()[0]]
                relevant_axes = f'{y}{z}'
            else:
                relevant_axes = 'xyz'
            if self._staggering_minimal_slicing_with_nontrivial_step(relevant_axes):
                # need to be careful to group stagger ops.
                simple_mode = False
        if simple_mode:
            # this is what happens "usually". Easier to read. Re-uses get_facecurl logic.
            for x in self.iter_component():
                y, z = YZ_FROM_X[x]
                centered_curl_x = self(f'{y}up_{z}up_facecurl_{here}', component=x)
                curls.append(centered_curl_x)
        else:
            # explicitly group stagger ops (including from facecurl), so padding occurs only once.
            for x in self.iter_component():
                y, z = YZ_FROM_X[x]
                ddy_vz = self(f'{y}up_{z}up_dd{y}dn_{here}_{z}')
                ddz_vy = self(f'{y}up_{z}up_dd{z}dn_{here}_{y}')
                curl_x = self.assign_component_coord(ddy_vz - ddz_vy, x)
                curls.append(curl_x)
        result = self.join_components(curls)
        return result



"""
File Purpose: fitting polynomials to values along a dimension.

# [TODO] write a test which compares linregt_var to ddt_var for a few vars.
#   they should be similar when data changes smoothly.
"""

from ..quantity_loader import QuantityLoader
from ...tools import (
    alias, alias_key_of,
    UNSET,
    xarray_coarsened_polyfit,
)

class PolyfitLoader(QuantityLoader):
    '''polyfit calculations. E.g. linear regression.
    For generic polyfitting see arr.polyfit or arr.pc.polyfit, where arr is an xarray.DataArray.
    '''
    # # # HELPER METHODS # # #
    def polyfit(self, array_or_var, coord, degree, window=UNSET, **kw):
        '''polyfit along coord. Might coarsen array, polyfit in each window, and concat results.

        array_or_var: xarray.DataArray, or str
            array to polyfit.
            str --> use array=self(array_or_var). (Note; **kw will NOT go to self.get(array_or_var))
        coord: str
            coordinate to polyfit along.
            If coord is not already a dimension, use array=promote_dim(array, coord).
        degree: int
            degree of polynomial to fit.
        window: UNSET, None, or int.
            UNSET --> use self.polyfit_window
            None --> don't use windowing; polyfit to the entire array along coord.
            int --> coarsen array along dim, using windows of this length,
                    then polyfit in each window, then concat results along coord.

        Pass additional kwargs to xarray_coarsened_polyfit;
            also use self.polyfit_kw as defaults (for any kwargs not provided here).

        returns an xarray.DataSet which is the result of polyfit.
        '''
        kw_use = {**self.polyfit_kw, **kw}
        window_self = kw_use.pop('window', None)
        if window is UNSET:
            window = window_self
        if isinstance(array_or_var, str):
            var = array_or_var
            array = self(var)
        else:
            array = array_or_var
        return xarray_coarsened_polyfit(array, coord, degree, window, **kw_use)

    # # # BEHAVIOR ATTRS # # #
    cls_behavior_attrs.register('polyfit_kw', default={'window': None, 'boundary': 'trim'})
    @property
    def polyfit_kw(self):
        '''kwargs to pass to self.polyfit(), other than array, coord, degree, and window.
        See polyfit_kw_key_aliases for a list of aliases to some of these kwargs, as attributes of self.

        getting self.polyfit_kw will also set keys to default values from aliases,
            e.g. polyfit_boundary has default of 'trim' --> if polyfit_kw['boundary'] not set, set it to 'trim'.
        '''
        try:
            result = self._polyfit_kw
        except AttributeError:
            result = dict()
            self._polyfit_kw = result
        if not getattr(self, '_inside_polyfit_kw_logic', False):
            with self.using(_inside_polyfit_kw_logic=True):
                for attr, key in self.polyfit_kw_key_aliases.items():
                    if key not in result:
                        val = getattr(self, attr)
                        if val is not UNSET:
                            result[key] = val
        return result
    @polyfit_kw.setter
    def polyfit_kw(self, v):
        '''set self.polyfit_kw.'''
        self._polyfit_kw = v   # defaults handled by getter.
    @polyfit_kw.deleter
    def polyfit_kw(self):
        '''delete self.polyfit_kw; restore defaults.'''
        del self._polyfit_kw   # defaults handled by getter.

    polyfit_kw_key_aliases = {  # attribute of self: key in self.polyfit_kw
            'polyfit_window': 'window',
            'polyfit_boundary': 'boundary',
            'polyfit_full': 'full',
            'polyfit_cov': 'cov',
            'polyfit_stddev': 'stddev',
            'polyfit_keep_coord': 'keep_coord',
    }

    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['window'] + self.polyfit_kw_key_aliases.keys() + any values from super().
        '''
        # [TODO] it's a little ambiguous what will happen if user enters polyfit_kw AND a polyfit_kw_key_alias.
        return ['window'] + list(self.polyfit_kw_key_aliases.keys()) + super()._extra_kw_for_quantity_loader_call

    polyfit_window = alias_key_of('polyfit_kw', 'window', default=None,
            doc='''When polyfitting, tells window size to use for coarsening arrays before fitting.
            E.g., polyfit_window=10 --> windows of length 10, polyfit in each window, concat results.''')

    window = alias('polyfit_window')  # deprecation warning: might become more generic than polyfit_window.

    polyfit_boundary = alias_key_of('polyfit_kw', 'boundary', default='trim',
            doc='''alias to self.polyfit_kw['boundary'].
            When polyfitting, tells how to handle boundaries when coarsening array.
            probably 'exact', 'trim', or 'pad'.''')

    polyfit_full = alias_key_of('polyfit_kw', 'full', default=UNSET,
            doc='''alias to self.polyfit_kw['full'].
            When polyfitting, tells whether to also return residuals, matrix rank and singular values.''')

    polyfit_cov = alias_key_of('polyfit_kw', 'cov', default=UNSET,
            doc='''alias to self.polyfit_kw['cov'].
            When polyfitting, tells whether to also return the covariance matrix.
            only used if self.polyfit_full=False.''')

    polyfit_stddev = alias_key_of('polyfit_kw', 'stddev', default=UNSET,
            doc='''alias to self.polyfit_kw['stddev'].
            When polyfitting, tells whether to also return the standard deviations of the coefficients.
            incompatible with self.polyfit_full=True.''')

    polyfit_keep_coord = alias_key_of('polyfit_kw', 'keep_coord', default=UNSET,
            doc='''alias to self.polyfit_kw['keep_coord'].
            When polyfitting, tells whether to keep some of the original coord values in result.
            probably 'left', 'middle', 'right', or False.''')

    # # # LOADABLE QUANTITIES # # #
    @known_pattern(r'linregt_(.+)', deps=[0], reduces_dims=['snap'])
    def get_linregt(self, var, *, _match=None):
        '''linear regression along 't' coord. result is an xarray.DataSet including 'polyfit_coefficients'.
        Equivalent to self.polyfit(var, 't', degree=1).

        behavior affected by self.polyfit_kw; see help(type(self).polyfit_kw) for details.
        '''
        here, = _match.groups()
        array = self(here)
        return self.polyfit(array, 't', degree=1)

    @known_pattern(r'slopet_(.+)', deps=[{0: 'linregt_{group0}'}])
    def get_slopet(self, var, *, _match=None):
        '''slope calculated from linear regression along 't' coord.
        self('slopet_var') == self('linregt_var)['polyfit_coefficients'].sel(degree=1).drop_vars('degree')

        behavior affected by self.polyfit_kw; see help(type(self).polyfit_kw) for details.
        '''
        here, = _match.groups()
        linregt = self(f'linregt_{here}')
        return linregt['polyfit_coefficients'].sel(degree=1).drop_vars('degree')  # don't keep the 'degree' coordinate.

    @known_pattern(r'growthrate_(.+)', deps=[{0: 'slopet_ln_abs_{group0}'}])
    def get_growthrate(self, var, *, _match=None):
        '''self('growthrate_var') --> exponential growth rate of var. i.e., slope of ln|var| vs 't'.
        self('growthrate_var') == self('slopet_ln_abs_var').

        behavior affected by self.polyfit_kw; see help(type(self).polyfit_kw) for details.
        '''
        here, = _match.groups()
        return self(f'slopet_ln_abs_{here}')

    @known_pattern(r'growth(_vs_)?k_(.+)', deps=[{1: 'growthrate_radfft_{group1}'}])
    def get_growthk(self, var, *, _match=None):
        '''self('growthk_var') --> exponential growth rate of radfft_var.
        i.e. for each k, gives slope of ln|radfft_var| vs 't'.
        self('growthk_var') == self('growthrate_radfft_var').
        
        Accepts strings starting with 'growth_vs_k_' or 'growthk_'.
        behavior affected by self.polyfit_kw; see help(type(self).polyfit_kw) for details.

        Assumes, but does not check, that fft_dims are spatial, only.
        '''
        _vs_str, here, = _match.groups()
        return self(f'growthrate_radfft_{here}')

    @known_pattern(r'growthfit_(.+)', deps=[{0: 'linregt_ln_abs_{group0}'}])
    def get_growthfit(self, var, *, _match=None):
        '''self('growthfit_var') --> linear regression along 't' coord, of ln|var|.

        result['polyfit_coefficients'].sel(degree=1).drop_vars('degree') gives the growth rate,
            because growth rate = slope of natural log of |var| vs 't'.

        result might also contain other variables, depending on self.polyfit_kw;
            might want to use polyfit_stddev=True, polyfit_cov=True or polyfit_full=True to get more info.
            see help(type(self).polyfit_kw) for more options.
        '''
        here, = _match.groups()
        return self(f'linregt_ln_abs_{here}')

    @known_pattern(r'growthfit(_vs_)?k_(.+)', deps=[{1: 'growthfit_radfft_{group1}'}])
    def get_growthfitk(self, var, *, _match=None):
        '''self('growthfitk_var') --> linear regression along 't' coord, of ln|radfft_var|.
        
        result['polyfit_coefficients'].sel(degree=1).drop_vars('degree') gives the growth rate,
            because growth rate = slope of natural log of |var| vs 't'.

        result might also contain other variables, depending on self.polyfit_kw;
            might want to use polyfit_stddev=True, polyfit_cov=True or polyfit_full=True to get more info.
            see help(type(self).polyfit_kw) for more options.

        Assumes, but does not check, that fft_dims are spatial, only.
        '''
        _vs_str, here, = _match.groups()
        return self(f'growthfit_radfft_{here}')



"""
Package Purpose: PlasmaCalculator for Copapic
"""
from .copapic_bases import CopapicBasesLoader
from .copapic_calculator import CopapicCalculator
from .copapic_dimensions import (
    CopapicDist, CopapicDistList,
    CopapicNeutral, CopapicNeutralList,
)
from .copapic_direct_loader import CopapicDirectLoader
from .copapic_input_deck import CopapicInputDeck
from .copapic_io_tools import (
    read_copapic_json_file,
    read_copapic_snaps_info
)



"""
File Purpose: electric field in Bifrost
"""
import numpy as np

from ...errors import FormulaMissingError
from ...quantities import QuantityLoader
from ...tools import alias, simple_property


''' --------------------- BifrostEfieldLoader --------------------- '''

class BifrostEfieldLoader(QuantityLoader):
    '''quantities related to electric field in Bifrost'''
    @known_var(dims=['snap'])
    def get_qjoule(self):
        '''joule heating (?). eta * |J|**2. Directly from Bifrost aux.
        qjoule in aux is stored in raw units.
            for more aux units see: https://ita-solar.github.io/Bifrost/aux_variables/
        '''
        if 'qjoule' not in self.directly_loadable_vars():
            raise FormulaMissingError('qjoule when not saved to aux')
        ufactor = self.u('energy_density time-1', convert_from='raw')
        return self.load_maindims_var_across_dims('qjoule', u=ufactor, dims=['snap'])

    @known_var(deps=['qjoule', 'mod2_J'])
    def get_eta(self):
        '''eta (scalar), such that E = -u x B + eta J + ....'''
        return self('qjoule') / self('mod2_J')

    cls_behavior_attrs.register('eta_hall_mode', default='best')
    ETA_HALL_MODE_OPTIONS = {
        'best': '''use 'aux' if possible, else |B|/(|qe| * self('ne', ne_mode='best'))''',
        'aux': '''use 'eta_hall' from aux''',
        'ne_best': '''|B|/(|qe| * self('ne', ne_mode='best'))''',
        'neq': '''|B| / (|qe| * self('ne_neq'))''',
        'table': '''|B| / (|qe| * self('ne_fromtable'))''',
        'ne': '''|B| / (|qe| * self('ne')) using present value of self.ne_mode.''',
    }
    eta_hall_mode = simple_property('_eta_hall_mode', default='best',
        validate_from='ETA_HALL_MODE_OPTIONS',
        doc='''tells where to get eta_hall from. See self.ETA_HALL_MODE_OPTIONS for details.''')

    eta_hall_mode_explicit = alias('eta_hall_mode',
        doc='''self.eta_hall_mode, but if 'best' convert to the explicit mode to use.''')
    @eta_hall_mode_explicit.getter
    def eta_hall_mode_explicit(self):
        mode = self.eta_hall_mode
        if mode == 'best':
            if 'eta_hall' in self.directly_loadable_vars():
                mode = 'aux'
            else:
                mode = 'ne_best'
        return mode

    _ETA_HALL_MODE_EXPLICIT_TO_NE_DEPS = {
        'aux': [],
        'ne_best': ['mod_B', 'abs_qe', ('ne', {'ne_mode': 'best'})],
        'neq': ['mod_B', 'abs_qe', 'ne_neq'],
        'table': ['mod_B', 'abs_qe', 'ne_fromtable'],
        'ne': ['mod_B', 'abs_qe', 'ne'],
    }
    @known_var(dims=['snap'],
               attr_deps=[('eta_hall_mode_explicit', '_ETA_HALL_MODE_EXPLICIT_TO_NE_DEPS')])
    def get_eta_hall(self):
        '''eta_hall (scalar), such that E = -u x B + eta_hall J x Bhat + ....
        eta_hall in aux is stored in raw(?) units ([TODO]-check!).
        eta_hall is equivalent to |B| / (ne * |qe|).

        self.eta_hall_mode controls whether to load from aux or use a formula;
            see self.ETA_HALL_MODE_OPTIONS for options & descriptions.

        result will have 'eta_hall_mode' attr telling the mode used.
            (if 'best' but aux exists, will say 'aux' instead of 'best'.
        '''
        mode = self.eta_hall_mode_explicit
        if mode == 'aux':
            if 'eta_hall' not in self.directly_loadable_vars():
                raise FormulaMissingError('eta_hall when not saved to aux')
            ufactor = self.u('e_field current_density-1', convert_from='raw')
            result = self.load_maindims_var_across_dims('eta_hall', u=ufactor, dims=['snap'])
        else:
            mode_to_ne_mode = {'ne_best': 'best', 'neq': 'neq', 'table': 'table', 'ne': self.ne_mode}
            result = self('(mod_B)/(abs_qe*ne)', ne_mode=mode_to_ne_mode[mode])
        return result.assign_attrs(eta_hall_mode=mode)

    @known_var(dims=['snap'])
    def get_eta_amb(self):
        '''eta_amb (scalar), such that E = -u x B - eta_amb (J x Bhat) x Bhat + ....
        eta_amb in aux is stored in raw(?) units ([TODO]-check!).
        '''
        if 'eta_amb' not in self.directly_loadable_vars():
            raise FormulaMissingError('eta_amb when not saved to aux')
        ufactor = self.u('e_field current_density-1', convert_from='raw')
        return self.load_maindims_var_across_dims('eta_amb', u=ufactor, dims=['snap'])

    @known_var(deps=['(SF_u)_cross_B'], aliases=['E_motional'])
    def get_E_uxB(self, *, _B=None):
        '''E_uxB = -u x B, the motional electric field. E = E_uxB + ...
        [EFF] for efficiency, can provide B if already known.
            CAUTION: if providing B, any missing components assumed to be 0.
        '''
        result = -self('(SF_u)_cross_B', _B=_B)
        # don't include fluid=SINGLE_FLUID in result.
        result = result.drop_vars('fluid')  # [TODO] handle 'fluid not in result coords' case?
        return result

    @known_var(deps=['eta', 'J'])
    def get_E_etaJ(self, *, _J=None):
        '''E_etaJ = eta * J. Goes into E = E_uxB + E_etaJ + ...
        [EFF] for efficiency, can provide J if already known.
        '''
        J = self('J') if _J is None else _J
        return self('eta') * J

    @known_var(deps=['eta_hall', 'J_cross_B'])
    def get_E_hall(self, *, _J=None, _B=None, _JxB=None, _mod_B=None):
        '''E_hall = eta_hall * J x B / |B|. Goes into E = E_uxB + E_hall + ...
        [EFF] for efficiency, can provide J, B, JxB, and/or _mod_B if already known.
            CAUTION: if providing vector, any missing components assumed to be 0.
        '''
        eta_hall = self('eta_hall')
        JxB = self('J_cross_B', _J=_J, _B=_B) if _JxB is None else _JxB
        mod_B = self('mod_B', _B=_B) if _mod_B is None else _mod_B
        return eta_hall * JxB / mod_B

    @known_var(deps=['eta_amb', 'J', 'B'])
    def get_E_amb(self, *, _J=None, _B=None, _JxB=None, _mod_B=None):
        '''E_amb = -eta_amb (J x B) x B / |B|**2. Goes into E = E_uxB + E_amb + ...
        [EFF] for efficiency, can provide J, B, JxB, and/or _mod_B if already known.
            CAUTION: if providing vector, any missing components assumed to be 0.
        '''
        eta_amb = self('eta_amb')
        JxB = self('J_cross_B', _J=_J, _B=_B) if _JxB is None else _JxB
        JxBxB = self('(J_cross_B)_cross_B', _val0=JxB, _B=_B)
        mod_B = self('mod_B', _B=_B) if _mod_B is None else _mod_B
        return -eta_amb * JxBxB / mod_B**2

    cls_behavior_attrs.register('E_from_aux', default=None)
    E_from_aux = simple_property('_E_from_aux', default=None,
        doc='''whether to load bifrost E from aux (ex ey ez).
        None --> True if ex ey ez all in aux, else False.
        True --> always load from ex ey ez aux (crash if not available)
        False --> never load from ex ey ez aux.''')

    E_from_aux_explicit = alias('E_from_aux',
        doc='''self.E_from_aux, but if None convert to the explicit mode to use.''')
    @E_from_aux_explicit.getter
    def E_from_aux_explicit(self):
        if self.E_from_aux is None:
            return all(x in self.directly_loadable_vars() for x in ['ex', 'ey', 'ez'])
        else:
            return self.E_from_aux

    @known_var(deps=['E_etaJ', 'E_hall', 'E_amb'])
    def get_E_u0(self, *, _J=None, _B=None, _JxB=None, _mod_B=None):
        '''E without motional electric field contribution. E_u0 = E_etaJ + E_hall + E_amb.
        if self.E_from_aux_explicit, instead computed via: E_u0 = E (from aux ex ey ez) - E_uxB.

        [EFF] for efficiency, can provide J, B, JxB, and/or _mod_B if already known.
            CAUTION: if providing vector, any missing components assumed to be 0.
        '''
        from_aux = self.E_from_aux_explicit
        if from_aux:
            result = self('E-E_uxB')
        else:
            with self.using(component=None):  # all 3 vector components here
                J = self('J') if _J is None else _J
                B = self('B') if _B is None else _B
                JxB = self('J_cross_B', _J=_J, _B=_B) if _JxB is None else _JxB
                mod_B = self('mod_B', _B=_B) if _mod_B is None else _mod_B
            E_etaJ = self('E_etaJ', _J=J)
            E_hall = self('E_hall', _J=J, _B=B, _JxB=JxB, _mod_B=mod_B)
            E_amb  = self('E_amb',  _J=J, _B=B, _JxB=JxB, _mod_B=mod_B)
            result = E_etaJ + E_hall + E_amb
        return result.assign_attrs(E_from_aux=from_aux)

    @known_var(deps=['E_uxB', 'E_u0'])
    def get_E(self):
        '''electric field from Bifrost. E = E_uxB + E_u0.
        loaded from aux ex ey ez if self.E_from_aux_explicit, else from E_uxB + E_u0.
        '''
        from_aux = self.E_from_aux_explicit
        if from_aux:
            result = self.load_maindims_var_across_dims('e', u='e_field', dims=['snap', 'component'])
        else:
            with self.using(component=None):  # all 3 vector components here
                B = self('B')
            E_uxB = self('E_uxB', _B=B)
            E_u0  = self('E_u0',  _B=B)
            result = E_uxB + E_u0
        return result.assign_attrs(E_from_aux=from_aux)

    @known_var(attr_deps=[('E_un0_type', '_E_UN0_TYPE_TO_DEPS')])
    def get_E_un0(self):
        '''electric field in u_n=0 frame.
        Result depends on self.E_un0_mode; see help(type(self).E_un0_mode) for details.
        Note: using E_un0_mode=None (the default) and self.assume_un='u',
            is equivalent to using E_un0_mode = 'un=u', which will just give self('E_u0').
        Else, E_un0 = E + u_n x B, if can get u_n (crash if can't get u_n).
        '''
        if self.E_un0_type == 'E+unxB':  # slightly more helpful error than super(); u_neutral will probably crash.
            try:
                un = self('u_neutral', component=None)  # check if can get u_n... probably cant...
            except NotImplementedError:
                errmsg = 'E_un0, when E_un0_mode=None, u_neutral unknown, and self.assume_un != "u".'
                raise FormulaMissingError(errmsg)
            else:
                result = self('E')
                if np.any(un != 0):
                    result = result + self('u_neutral_cross_B', _u_neutral=un)
        else:
            result = super().get_E_un0()
        return result

    @property   # [TODO] should be a known var instead.
    def E_un0_type(self):
        '''string telling method that will be used to get E_un0. Based on self.E_un0_mode and assume_un.
        possible results:
            'nan' <--> will crash. (e.g. this occurs if E_un0_mode = 'E0_perpmodB_min')
            'E+unxB' <--> self('E') + self('u_neutral_cross_B')
            'E_u0' <--> self('E_u0')
            'E+uxB' <--> self('E') + self('u_cross_B')
            'E' <--> self('E')
            'E0_perpB' <--> self('E0_un0_perpB')
        '''
        if self.E_un0_mode is None and self.assume_un == 'u':
            return 'E_u0'
        else:
            return super().get_E_un0_type()



"""
File Purpose: calculating plasma parameters, e.g. plasma beta, plasma frequency
"""
import numpy as np

from .quantity_loader import QuantityLoader
from ..dimensions import ELECTRON, IONS
from ..tools import xarray_sum

class PlasmaParametersLoader(QuantityLoader):
    '''plasma parameters, e.g. plasma beta, plasma frequency'''
    # # # PLASMA BETA # # #
    @known_var(deps=['P', 'mod_B'])
    def get_beta(self):
        '''plasma beta. beta = (pressure / magnetic pressure) = (P / (B^2 / (2 mu0)))'''
        return self('P') / (self('mod_B')**2 / (2 * self.u('mu0')))

    # # # GYROFREQUENCY # # #
    @known_var(deps=['q', 'mod_B', 'm'])
    def get_sgyrof(self):
        '''signed gyrofrequency. sgyrof == q |B| / m == charge * |B| / mass. Negative when charge < 0.'''
        return self('mod_B') * (self('q') / self('m'))
        # ^note: grouped (q/m) for efficiency (both are probably constant across maindims)
        #    and because both might be small numbers --> helps avoid getting close to float32 limits.

    @known_var(deps=['abs_sgyrof'], aliases=['cyclof', 'omega_c'])
    def get_gyrof(self):
        '''(unsigned) gyrofrequency. gyrof == |sgyrof| == |q| |B| / m == |charge| * |B| / mass.'''
        return self('abs_sgyrof')

    # # # MAGNETIZATION PARAMETER (KAPPA) # # #
    @known_var(deps=['nusn', 'sgyrof'])
    def get_skappa(self):
        '''signed kappa (magnetization parameter). skappa = sgyrof / nusn. Negative when charge < 0.
        skappa = gyrofrequency / collision frequency of self.fluid with neutrals.
        gyrofrequency == q * |B| / (mass * nusn). 
        ''' 
        return self('sgyrof') / self('nusn')

    @known_var(deps=['abs_skappa'])
    def get_kappa(self):
        '''(unsigned) kappa (magnetization parameter). kappa = |skappa| == |gyrof| / nusn.
        kappa = |gyrofrequency| / collision frequency of self.fluid with neutrals.
        '''
        return self('abs_skappa')

    # # # PLAMSA FREQUENCY # # #
    @known_var(deps=['n', 'abs_q', 'm'])
    def get_wplasma(self):
        '''"plasma frequency" for self.fluid. wplasma = sqrt(n q^2 / (m epsilon0))
        This is analogous to the "true" plasma frequency of Langmuir oscillations,
            which is calculated using the same formula but applied to electrons.
        wplasma is equivalent to wplasmae when self.fluid is electrons.
        '''
        return self('n')**0.5 * self('abs_q') / (self('m') * self.u('eps0'))**0.5

    @known_var(deps=['wplasma'], aliases=['wpe', 'omega_pe'], ignores_dims=['fluid'])
    def get_wplasmae(self):
        '''electron plasma frequency; Langmuir oscillations. wpe = sqrt(ne qe^2 / (me epsilon0))'''
        return self('wplasma', fluid=ELECTRON)

    # # # DEBYE LENGTH # # #
    @known_var(deps=['ldebye2'])
    def get_ldebye(self):
        '''Debye length (of self.fluid). ldebye = sqrt(epsilon0 kB T / (n q^2))'''
        return self('ldebye2')**0.5

    @known_var(deps=['Tjoule', 'n', 'abs_q'])
    def get_ldebye2(self):
        '''squared Debye length (of self.fluid). ldebye2 = epsilon0 kB T / (n q^2)'''
        return self.u('eps0') * self('Tjoule') / (self('n') * self('abs_q')**2)
        
    @known_var(deps=['ldebye2'], reduces_dims=['fluid'])
    def get_ldebye_subset(self):
        '''"total" Debye length; ldebye_subset = sqrt(epsilon0 kB / sum_fluids(n q^2 / (kB T)))
        sum is taken over the fluids in self.fluid.
        Equivalent: sqrt( 1 / sum_fluids(1/ldebye^2) )
        '''
        return xarray_sum(1/self('ldebye2'), dim='fluid')**-0.5

    @known_var(deps=['ldebye_subset'], ignores_dims=['fluid'])
    def get_ldebye_total(self):
        '''total Debye length for all fluids; ldebye_total = sqrt(epsilon0 kB / sum_fluids(n q^2 / (kB T)))
        sum is taken over all the fluids in self.fluids.
        Equivalent: sqrt( 1 / sum_fluids(1/ldebye^2) )
        '''
        return self('ldebye_subset', fluid=None)  # fluid=None --> use all fluids.

    # # # MEAN FREE PATH # # #
    @known_var(deps=['vtherm', 'nusn'], aliases=['lmfp'])
    def get_mean_free_path(self):
        '''collisional mean free path. lmfp = vtherm / nusn = thermal velocity / collision frequency.'''
        return self('vtherm') / self('nusn')

    # # # THERMAL VELOCITY # # #
    @known_var(deps=['Tjoule', 'm'], aliases=['vth', 'vthermal'])
    def get_vtherm(self):
        '''thermal velocity. vtherm = sqrt(kB T / m)'''
        return (self('Tjoule') / self('m'))**0.5

    @known_setter
    def set_vtherm(self, value, **kw):
        '''set thermal velocity, by setting T.
        vtherm = sqrt(kB T / m) --> set T to (m vtherm^2 / kB).
        '''
        T = self('m').values * value**2 / self.u('kB')
        self.set('T', T, **kw)

    @known_var(deps=['T_n', 'm_n'], aliases=['vth_n', 'vthermal_n',
                                'vth_neutral', 'vtherm_neutral', 'vthermal_neutral'])
    def get_vtherm_n(self):
        '''thermal velocity for neutrals. vtherm_n = sqrt(kB T_n / m_n)'''
        return (self('kB*T_n') / self('m_n'))**0.5

    # # # SOUND SPEED # # #
    @known_var(deps=['csound2'], aliases=['cs'])
    def get_csound(self):
        '''sound speed. csound = sqrt(gamma * P / r)'''
        return self('csound2')**0.5

    @known_var(deps=['gamma', 'P', 'r'], aliases=['cs2'])
    def get_csound2(self):
        '''sound speed squared. csound2 = gamma P / r.'''
        return self('gamma') * self('P') / self('r')

    # # # ALFVEN SPEED # # #
    @known_var(deps=['va2'], aliases=['va'])
    def get_valfven(self):
        '''Alfven speed. valfven = |B| / sqrt(mu0 * r)'''
        return self('va2')**0.5

    @known_var(deps=['mod2_B', 'r'], aliases=['va2'])
    def get_valfven2(self):
        '''Alfven speed squared. valfven2 = |B|^2 / (mu0 * r)'''
        return self('mod2_B') / (self.u('mu0') * self('r'))

    # # # ni/ne # # #
    @known_var(deps=['n'], ignores_dims=['fluid'])
    def get_niefrac(self):
        '''ion-to-electron density ratio. niefrac = ni / ne.
        Result always corresponds to fluid=IONS, regardless of current self.fluid.
        '''
        return self('n', fluid=IONS) / self('n', fluid=ELECTRON)

    # # # ROSENBERG CRITERION # # #
    # [TODO] this should move to a separate file, it's too specific for "generic plasma parameters" loader.
    @known_var(deps=['nusn', 'wplasma'], aliases=['rosenberg'])
    def get_rosenberg_qn(self):
        '''Rosenberg criterion for quasineutrality, for each fluid: rosenberg_qn = (nusn / wplasma)^2.
        quasineutrality is "reasonable" (during farley-buneman analysis) iff rosenberg_qn << 1 for ions.
        (Intuitively: quasineutrality reasonable iff 'collisions much slower than plasma oscillations')

        If multiple ions, consider using self('rosenberg_multi') instead of one criterion per ion.

        see Rosenberg 1998, equation 17, for details.
        '''
        return self('nusn/wplasma')**2

    @known_var(deps=['nusn', 'abs_q', 'm'], aliases=['n_at_lmfp_eq_ldebye', 'n_at_wplasma_eq_nusn'])
    def get_rosenberg_n(self):
        '''n such that rosenberg_qn == 1, for each fluid.
        Equivalent: rosenberg_qn * n.
        To satisfy Rosenberg criterion, need rosenberg_qn << 1.
        rosenberg_qn = (nusn / wplasma)^2 = nusn^2 * m * eps0 / (n * q^2),
            which is proportional to 1 / n
            --> n is "good" if n >> rosenberg_n.

        Note: this also happens to be the solution to lmfp == ldebye...
            lmfp = (vtherm/nusn) = ((kB T / m)^0.5 / nusn) == (epsilon0 kB T / (n q^2))^0.5 = ldebye
            --> (nusn^-2 m^-1) == (epsilon0 n^-1 q^-2)
            --> n == epsilon0 nusn^2 m / q^2.
        '''
        return self('nusn')**2 * (self('m') * self.u('eps0') / self('abs_q')**2)

    @known_var(deps=['rosenberg_n', 'n'])
    def get_rosenberg_n_margin(self):
        '''n / rosenberg_n. margin of safety for rosenberg criterion. "safe" if margin is large.'''
        return self('n') / self('rosenberg_n')

    @known_var(deps=['wplasma'], ignores_dims=['fluid'])
    def get_rosenberg_multi_wplasma(self):
        '''plasma frequency for rosenberg criterion with multiple ions.
        wplasma_multi^2 = 1 / (1 / wplasma1^2 + 1 / wplasma2^2 + ...),
            with sum across all ions in self('wplasma', fluids=None)
        see rosenberg_multi for details.
        '''
        wplasma_ions = self('wplasma', fluid=IONS)
        denom = xarray_sum(1 / wplasma_ions**2, dim='fluid')
        return denom**-0.5

    @known_var(deps=['nusn'], ignores_dims=['fluid'])
    def get_rosenberg_multi_nusn(self):
        '''collision frequency for rosenberg criterion with multiple ions.
        nusn_multi = nu1 * weight1 + nu2 * weight2 + ... / (weight1 + weight2 + ...),
            where weightj = wplasma_1^2 * wplasma_2^2 * ... / wplasma_j^2.
        see rosenberg_multi for details.
        '''
        wplasma_ions = self('wplasma', fluid=IONS)
        nusn_ions = self('nusn', fluid=IONS)
        # compute weights in log-space:
        # log(weightj) = log(wplasma_1^2) + log(wplasma_2^2) + ... - log(wplasma_j^2)
        log_wplasma2 = np.log10(wplasma_ions**2)
        log_weight_prod = xarray_sum(log_wplasma2, dim='fluid')
        log_weight_j = log_weight_prod - log_wplasma2
        # compute coefficients in log space:
        # coeff_j = weightj / sum(weights)
        # log(coeff_j) = log(weightj) - log(sum(weights))
        weight_j = 10**log_weight_j
        log_sum_weights = np.log10(xarray_sum(weight_j, dim='fluid'))
        log_coeff_j = log_weight_j - log_sum_weights
        coeff_j = 10**log_coeff_j
        terms = nusn_ions * coeff_j
        return xarray_sum(terms, dim='fluid')

    @known_var(deps=['rosenberg_multi_wplasma', 'rosenberg_multi_nusn'])
    def get_rosenberg_multi(self):
        '''rosenberg criterion for multiple ions: rosenberg_multi = (nusn_multi / wplasma_multi)^2.
        quasineutrality is "reasonable" (during farley-buneman analysis) iff rosenberg_multi << 1.
        This criterion should be more accurate than using rosenberg_qn for each ion, separately.

        (Below uses '...' to denote "terms independent of wplasma_i and nu_in".)
        Rosenberg 1998 equation 17, derived from equation 14, assumed only 1 ion.
        However, equation 14 comes from 13, where wplasma_i and nu_in only appear as (Ai/wplasma_i^2).
            where Ai = omega * (omega + i nu_in) + ...   (see equation 15).
            equation 13 looks like: ... + (Ai/wplasma_i^2) = 0
        From equation 4 I infer the dispersion relation with multiple ions will be like:
            ... + sum_i (Ai/wplasma_i^2) = 0.
        Expanding this sum reveals that it can be expressed in the same algebraic form as with 1 ion:
            ... + Am/wplasma_m^2 = 0, when:
            wplasma_m^2 = 1 / (1/wplasma_1^2 + 1/wplasma_2^2 + ...),
            Am = omega * (omega + i nu_mn) + ...
            nu_mn = nu1 * weight1 + nu2 * weight2 + ... / (weight1 + weight2 + ...),
                where weightj = wplasma_1^2 * wplasma_2^2 * ... / wplasma_j^2.
        Thus, the remaining steps done by Rosenberg to derive the criterion should apply in the same way,
            and for multiple ions we can conclude the criterion is (nu_mn / wplasma_m)^2 << 1.


        Okay, actually, in practice, it's probably not useful to use this...
            it mostly just selects the least-dense ion, due to the 1/wplasma^2 scaling.
        Also, the criterion means "instability gets dampened when collisions are faster than plasma oscillations"
            But for multiple ions, it's really more like, the instability for THAT ion gets dampened...
            (Also, I think one of the assumptions in the paper probably fails when wplasma is small...)
        '''
        return self('rosenberg_multi_nusn/rosenberg_multi_wplasma')**2



"""
File Purpose: collision frequency calculations
"""
import numpy as np
import xarray as xr

from .collisions_cross_sections_loader import CollisionsCrossSectionsLoader
from ...errors import (
    QuantInfoMissingError, CollisionsModeError,
    FluidValueError,
)
from ...tools import simple_property, UNSET
from ...defaults import DEFAULTS


class CollisionsLoader(CollisionsCrossSectionsLoader):
    '''collision frequency calculations, for collisions related to momentum / energy transfer.
    (Collisions between the same species, e.g. electron-electron or proton-proton, are assumed to be 0)'''
    cls_behavior_attrs.register('collisions_mode', default='best')

    COLLISIONS_MODE_OPTIONS = {
        'best':
            '''Use coulomb for charge-charge, fromtable if table exists, maxwell otherwise.
            For charge-neutral, if table doesn't exist, collisions_cross_section will give 0.''',
        'table':
            '''Use coulomb for charge-charge, else fromtable.
            For charge-neutral, if table doesn't exist, raise QuantInfoMissingError.''',
        'maxwell':
            '''Use coulomb for charge-charge, maxwell otherwise.
            Does not affect collisions_cross_section.''',
        'neutral_only':
            '''Use 0 for charge-charge, fromtable if table exists, maxwell otherwise.
            For charge-neutral, if table doesn't exist, collisions_cross_section will give 0.''',
        'table_only':
            '''Use 0 for charge-charge, else fromtable.
            For charge-neutral, if table doesn't exist, raise QuantInfoMissingError.''',
    }

    collisions_mode = simple_property('_collisions_mode', default='best', validate_from='COLLISIONS_MODE_OPTIONS',
        doc='''str, the collisions mode to use when getting nusj. See COLLISIONS_MODE_OPTIONS for details.
        Note that you can always calculate collisions using a specific formula with the appropriate var,
            regardless of collisions_mode; e.g., nusj_maxwell will always use maxwell formula.''')

    # # # ALL COLLISIONS (DISPATCHER) # # #
    @known_var(load_across_dims=['fluid', 'jfluid'])
    def get_collision_type(self):
        '''collision type between self.fluid and self.jfluid.
        result depends on fluids and on self.collisions_mode.
            '0' <--> fluid and jfluid are the same fluid.
            'coulomb' <--> do coulomb collisions
            'maxwell' <--> do maxwell collisions
            'fromtable' <--> get collision cross section from a table
        '''
        fs = self.fluid
        fj = self.jfluid
        mode = self.collisions_mode
        if fs == fj:  # same-same collisions
            result = '0'
        # hard-code the options list here in case subclass provides more keys in COLLISIONS_MODE_OPTIONS;
        # the implementation below assumes collisions_mode is one of these 5 options.
        elif mode not in ('best', 'table', 'maxwell', 'neutral_only', 'table_only'):
            errmsg = f'{type(self).__name__}.get_collision_type not implemented when collisions_mode = {mode!r}'
            raise CollisionsModeError(errmsg)
        elif fs.q != 0 and fj.q != 0:  # coulomb collisions
            if mode in ('neutral_only', 'table_only'):
                result = '0'
            else:
                result = 'coulomb'
        else:  # charge-neutral or neutral-neutral collisions.
            if mode in ('table', 'table_only'):
                result = 'fromtable'
            elif mode == 'maxwell':
                result = 'maxwell'
            else:
                assert mode in ('best', 'neutral_only')
                table_exists = ((fs, fj) in self.collisions_cross_mapping)
                if table_exists:
                    result = 'fromtable'
                else:
                    result = 'maxwell'
        return xr.DataArray(result)

    COLLISION_TYPE_TO_VAR = {  # used in self.get_nusj.
        # Subclasses might copy this dict and update it with more types.
        '0': '0',
        'coulomb': 'nusj_coulomb',
        'maxwell': 'nusj_maxwell',
        'fromtable': 'nusj_fromtable'
    }

    @known_var(partition_across_dim=('fluid', 'collision_type'), partition_deps='COLLISION_TYPE_TO_VAR')
    def get__nusj_singlej(self, *, collision_type):
        '''_nusj_singlej = collision frequency between self.fluid and a single jfluid.
        Implementation detail... may be removed later. Use self('nusj') instead.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if self.jfluid_is_iterable():
            raise FluidValueError('get__nusj_singlej should only be used for a single jfluid.')
        try:
            var = self.COLLISION_TYPE_TO_VAR[collision_type]
        except KeyError:
            raise ValueError(f'unknown collision_type: {collision_type!r}')
        else:
            return self(var)

    @known_var(load_across_dims=['jfluid'], deps=['_nusj_singlej'])
    def get_nusj(self):
        '''collision frequency.
        "frequency for one particle of s (self.fluid) to collide with any of j (self.jfluid)."
        Uses the appropriate formula based on self.collisions_mode, and self.fluid & self.jfluid.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self('_nusj_singlej')

    # # # COULOMB COLLISIONS # # #
    @known_var(deps=['T', 'n'])
    def get_coulomb_logarithm(self):
        '''Coulomb logarithm, ln(Lambda). Used in coulomb collisions.
        result = ln(Lambda) = 23.0 + 1.5 ln(Te [K] / 10^6) - 0.5 ln(ne [m^-3] / 10^12)
        '''
        electron = self.fluids.get_electron()  # get the one and only electron fluid.
        with self.using(fluid=electron, units='si'):
            T = self('T')
            ne = self('n')
        return 23.0 + 1.5 * np.log(T / 1e6) - 0.5 * np.log(ne / 1e12)

    @known_var(aliases=['nusj_cl'], deps=['coulomb_logarithm', 'T', 'm', 'q', 'n', 'm_sj', 'T_sj', 'delta_sj'])
    def get_nusj_coulomb(self):
        '''coulomb collision frequency.
        "frequency for one particle of s (self.fluid) to collide with any of j (self.jfluid)."
        This is a good model for nusj when:
            - s & j are both charged species.

        nusj_coulomb = (1 Hz) * (1.7/20) * (ln(Lambda))
                        * (m_p/m_s) * (m_sj/m_p)^(1/2)
                        * (T_sj / Kelvin)^(-3/2)
                        * (q_s/q_e)^2 * (q_j/q_e)^2
                        * (n_j / (10^6 meters^-3)), where:
            ln(Lambda) = Coulomb logarithm (see self.help('coulomb_logarithm') for details)
            n_j = number density of j (self.jfluid).
            m_sj = weighted mass = m_s * m_j / (m_s + m_j)
            T_sj = mass-weighted temperature = (m_s * T_j + m_j * T_s) / (m_s + m_j)
            m_p = mass of proton
            q_e = charge of electron
            m_s, m_j = mass of s, j
            T_s, T_j = temperature of s, j.
            q_s, q_j = charge of s, j
        '''
        logcoul = self('coulomb_logarithm')  # [dimensionless]
        m_p = self.u('m_p', 'si')
        q_e = self.u('q_e', 'si')
        with self.using(units='si'):
            m_s = self('m')
            q_s = self('q')
            T_s = self('T')
            m_j = self.getj('m')
            q_j = self.getj('q')
            T_j = self.getj('T')
            n_j = self.getj('n')
        m_sj = self('m_sj', _m_s=m_s, _m_j=m_j)
        T_sj = self('T_sj', _m_s=m_s, _T_s=T_s, _m_j=m_j, _T_j=T_j)
        Z_s = q_s / q_e
        Z_j = q_j / q_e
        result_si = (1.7 / 20) * logcoul * \
                    (m_p / m_s) * np.sqrt(m_sj / m_p) * \
                    T_sj**(-3/2) * \
                    Z_s**2 * Z_j**2 * \
                    (n_j / 1e6)
        result = result_si * self.u('Hz', self.units, convert_from='si')
        result = result * (1 - self('delta_sj'))   # exclude same-same collisions
        return result

    # # # MAXWELL COLLISIONS # # #
    @known_var(aliases=['nusj_mx'], deps=['m', 'n', 'delta_sj'])
    def get_nusj_maxwell(self):
        '''maxwell collision frequency.
        "frequency for one particle of s (self.fluid) to collide with any of j (self.jfluid)."
        This is a good model for nusj when:
            - s or j is neutral hydrogen.
                (assumed by the implementation here; could be improved in the future)
            - the charged species is a heavy ion.
                (collisions between [e-,H] and [H+,H] behave differently.)

        nusj_maxwell = (1 Hz) * (1.95856) * n_j * sqrt(
                        (alpha_n * q_e^2 / eps0) * (m_j / (m_s * (m_s + m_j)))
                        ), where:
            n_j = number density of j [si units, i.e. meters^-3]
            alpha_n = polarizability of neutral hydrogen. [si units, i.e. meters^3]
            q_e = charge of electron. [si units, i.e. C]
            eps0 = permittivity of free space; standard definition for eps0. [si units]
            m_s, m_j = mass of s, j. [si units, i.e. kg]

        (formula from Oppenheim+2020, Appendix A.)
        '''
        ## constants. 
        CONST_MULT = 1.95856  # factor in front.
        CONST_ALPHA_N = 6.67e-31  # [m^3]    #polarizability for Hydrogen   #(should be different for different species)
        qe = self.u('qe', 'si')    # [C]
        eps0 = self.u('eps0', 'si')  # [kg^-1 m^-3 s^4 (C^2 s^-2)] #epsilon0, standard definition
        amu = self.u('amu', 'si')   # [kg]   one atomic mass unit, in kg
        CONST_RATIO = (qe / amu) * (qe / eps0) * CONST_ALPHA_N   # [s^-2]   # parenthesis to avoid floating point errors
        # -- units of CONST_RATIO: [C^2 kg^-1 [eps0]^-1 m^3] == [C^2 kg^-1 (kg^1 m^3 s^-2 C^-2) m^-3] == [s^-2]
        ## values of other variables in formula
        with self.using(units='si'):
            m_s = self('m') / amu    # [amu]; mass of species s in amu
            m_j = self.getj('m') / amu   # [amu]; mass of species j in amu
            n_j = self.getj('n')
        ## the actual calculation
        result_si = CONST_MULT * n_j * np.sqrt(CONST_RATIO * m_j / (m_s * (m_s + m_j)))
        result = result_si * self.u('Hz', self.units, convert_from='si')
        result = result * (1 - self('delta_sj'))   # exclude same-same collisions
        return result

    # # # FROMTABLE COLLISIONS (i.e., cross section from a table) # # #
    @known_var(deps=['collisions_cross_section', 'T', 'm', 'n', 'm_sj', 'T_sj', 'delta_sj'])
    def get_nusj_fromtable(self):
        '''collision frequency from a table.
        "frequency for one particle of s (self.fluid) to collide with any of j (self.jfluid)."
        Requires that the relevant CrossTable(s) are provided, in self.collisions_cross_mapping.
        
        nusj_fromtable = n_j * (m_j / (m_s + m_j)) * C(T_sj) * sqrt(8 * kB * T_sj / (pi * m_sj)), where:
            n_j = number density of j (self.jfluid)
            m_sj = weighted mass = m_s * m_j / (m_s + m_j)
            T_sj = mass-weighted temperature = (m_s * T_j + m_j * T_s) / (m_s + m_j)
            m_s, m_j = mass of s, j
            T_s, T_j = temperature of s, j
            C(T_sj) = collisions_cross_section between s & j at temperature T_sj
            kB = Boltzmann constant
        '''
        cross = self('collisions_cross_section')
        m_s = self('m')
        T_s = self('T')
        m_j = self.getj('m')
        T_j = self.getj('T')
        n_j = self.getj('n')
        m_sj = self('m_sj', _m_s=m_s, _m_j=m_j)
        T_sj = self('T_sj', _m_s=m_s, _T_s=T_s, _m_j=m_j, _T_j=T_j)
        Tsj_speed = np.sqrt((8 * self.u('kB')/(np.pi * m_sj)) * T_sj)
        result = n_j * (m_j / (m_s + m_j)) * cross * Tsj_speed
        return result

    def _handle_missing_collisions_crosstab(self):
        '''handles the case of a missing collisions cross table between self.fluid & self.jfluid.
        depending on self.collisions_mode, this either returns 0, or raises a QuantInfoMissingError
        see self.COLLISIONS_MODE_OPTIONS for details.
        '''
        mode = self.collisions_mode
        RETURN_0_OPTIONS = ('best', 'neutral_only')
        RAISE_ERROR_OPTIONS = ('table', 'table_only', 'maxwell')
        if mode in RETURN_0_OPTIONS:
            return self('0')   # 0, as an xarray
        elif mode in RAISE_ERROR_OPTIONS:
            errmsg = ('missing a required cross table. This error can be fixed via set_collisions_crosstab,\n'
                     f'or ignored by switching collisions_mode from {mode!r} to one of {RETURN_0_OPTIONS}.')
            raise QuantInfoMissingError(errmsg)
        else:
            raise CollisionsModeError(f'unknown collisions_mode: {mode!r}')

    # # # COLLISIONS HELPER VARS # # #
    @known_var(deps=['m'])
    def get_m_sj(self, *, _m_s=UNSET, _m_j=UNSET):
        '''weighted mass. m_sj = m_s * m_j / (m_s + m_j).
        s = self.fluid; j = self.jfluid.
        [EFF] for efficiency, can provide m_s and/or m_j, if already known.
        '''
        m_s = self('m') if _m_s is UNSET else _m_s
        m_j = self.getj('m') if _m_j is UNSET else _m_j
        # parenthesis to avoid floating point errors:
        m_sum = m_s + m_j
        m_sj = m_s * (m_j / m_sum)
        return m_sj

    @known_var(deps=['m', 'T'])
    def get_T_sj(self, *, _m_s=UNSET, _T_s=UNSET, _m_j=UNSET, _T_j=UNSET):
        '''mass-weighted temperature. T_sj = (m_s * T_j + m_j * T_s) / (m_s + m_j).
        s = self.fluid; j = self.jfluid.
        [EFF] for efficiency, can provide any of m_s, T_s, m_j, T_j, if already known.
        '''
        m_s = self('m') if _m_s is UNSET else _m_s
        T_s = self('T') if _T_s is UNSET else _T_s
        m_j = self.getj('m') if _m_j is UNSET else _m_j
        T_j = self.getj('T') if _T_j is UNSET else _T_j
        # parenthesis to avoid floating point errors:
        m_sum = m_s + m_j
        T_sj = (m_s / m_sum) * T_j + (m_j / m_sum) * T_s
        return T_sj

    @known_var(load_across_dims=['fluid', 'jfluid'])
    def get_delta_sj(self):
        '''kronecker delta function between self.fluid and self.jfluid.
        1 where self.fluid == self.jfluid; 0 otherwise.
        Useful e.g. to avoid counting collisions between the same species.
        '''
        # [TODO][EFF] faster to avoid calculating same-species collisions in the first place,
        #     instead of calculating them and then multiplying by (1 - delta_sj)?
        if self.fluid == self.jfluid:
            return self('1')  # 1, as an xarray
        else:
            return self('0')  # 0, as an xarray
